# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from collections import defaultdict
from functools import cached_property
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

import earthkit.data as ekd
from anemoi.utils.checkpoints import load_metadata
from earthkit.data.utils.dates import to_datetime

from anemoi.inference.forcings import Forcings
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date
from anemoi.inference.types import State

from .metadata import Metadata

LOG = logging.getLogger(__name__)


def _download_huggingfacehub(huggingface_config: Any) -> str:
    """Download model from huggingface.

    Parameters
    ----------
    huggingface_config : dict or str
        Configuration for downloading from huggingface.

    Returns
    -------
    str
        Path to the downloaded model.
    """
    try:
        from huggingface_hub import hf_hub_download
        from huggingface_hub import snapshot_download
    except ImportError as e:
        raise ImportError("Could not import `huggingface_hub`, please run `pip install huggingface_hub`.") from e

    if isinstance(huggingface_config, str):
        huggingface_config = {"repo_id": huggingface_config}

    if "filename" in huggingface_config:
        return str(hf_hub_download(**huggingface_config))

    repo_path = Path(snapshot_download(**huggingface_config))
    ckpt_files = list(repo_path.glob("*.ckpt"))

    if len(ckpt_files) == 1:
        return str(ckpt_files[0])
    else:
        raise ValueError(
            f"None or Multiple ckpt files found in repo, {ckpt_files}.\nCannot pick one to load, please specify `filename`."
        )


class Checkpoint:
    """Represents an inference checkpoint."""

    def __init__(
        self,
        source: Union[str, Metadata, Dict[str, Any]],
        *,
        patch_metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialize the Checkpoint.

        Parameters
        ----------
        path : str
            The path to the checkpoint.
        patch_metadata : Optional[Dict[str, Any]], optional
            Metadata to patch the checkpoint with, by default None.
        """
        self._source = source
        self.patch_metadata = patch_metadata

    def __repr__(self) -> str:
        """Represent the Checkpoint as a string.

        Returns
        -------
        str
            String representation of the Checkpoint.
        """
        return f"Checkpoint({self.path})"

    @cached_property
    def path(self) -> str:
        """Get the path to the checkpoint."""
        import json

        try:
            path = json.loads(self._source)
        except Exception:
            path = self._source

        if isinstance(path, (Path, str)):
            return str(path)
        elif isinstance(path, dict):
            if "huggingface" in path:
                return _download_huggingfacehub(path["huggingface"])
            pass
        raise TypeError(f"Cannot parse model path: {path}. It must be a path or dict")

    @cached_property
    def _metadata(self) -> Metadata:
        """Get the metadata."""

        if isinstance(self._source, Metadata):
            return self._source

        try:
            result = Metadata(*load_metadata(self.path, supporting_arrays=True))
        except Exception as e:
            LOG.warning("Version does not support `supporting_arrays` (%s)", e)
            result = Metadata(load_metadata(self.path))

        if self.patch_metadata:
            LOG.warning("Patching metadata with %r", self.patch_metadata)
            result.patch(self.patch_metadata)

        return result

    ###########################################################################
    # Forwards used by the runner
    # We do not want to expose the metadata object directly
    # We do not use `getattr` to avoid exposing all methods and make debugging
    # easier
    ###########################################################################

    @property
    def timestep(self) -> Any:
        """Get the timestep."""
        return self._metadata.timestep

    @property
    def precision(self) -> Any:
        """Get the precision."""
        return self._metadata.precision

    @property
    def number_of_grid_points(self) -> Any:
        """Get the number of grid points."""
        return self._metadata.number_of_grid_points

    @property
    def number_of_input_features(self) -> Any:
        """Get the number of input features."""
        return self._metadata.number_of_input_features

    @property
    def variable_to_input_tensor_index(self) -> Any:
        """Get the variable to input tensor index."""
        return self._metadata.variable_to_input_tensor_index

    @property
    def model_computed_variables(self) -> Any:
        """Get the model computed variables."""
        return self._metadata.model_computed_variables

    @property
    def typed_variables(self) -> Any:
        """Get the typed variables."""
        return self._metadata.typed_variables

    @property
    def diagnostic_variables(self) -> Any:
        """Get the diagnostic variables."""
        return self._metadata.diagnostic_variables

    @property
    def prognostic_variables(self) -> Any:
        """Get the prognostic variables."""
        return self._metadata.prognostic_variables

    @property
    def prognostic_output_mask(self) -> Any:
        """Get the prognostic output mask."""
        return self._metadata.prognostic_output_mask

    @property
    def prognostic_input_mask(self) -> Any:
        """Get the prognostic input mask."""
        return self._metadata.prognostic_input_mask

    @property
    def output_tensor_index_to_variable(self) -> Any:
        """Get the output tensor index to variable."""
        return self._metadata.output_tensor_index_to_variable

    @property
    def accumulations(self) -> Any:
        """Get the accumulations."""
        return self._metadata.accumulations

    @property
    def latitudes(self) -> Any:
        """Get the latitudes."""
        return self._metadata.latitudes

    @property
    def longitudes(self) -> Any:
        """Get the longitudes."""
        return self._metadata.longitudes

    @property
    def grid_points_mask(self) -> Any:
        """Get the grid points mask."""
        return self._metadata.grid_points_mask

    @cached_property
    def sources(self) -> List["SourceCheckpoint"]:
        """Get the sources."""
        return [SourceCheckpoint(self, _) for _ in self._metadata.sources(self.path)]

    def default_namer(self, *args: Any, **kwargs: Any) -> Callable[[ekd.Field, Any], str]:
        """Return a callable that can be used to name fields.

        Parameters
        ----------
        *args : Any
            Additional arguments.

        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Callable
            The namer that was used to create the training dataset.
        """
        return self._metadata.default_namer(*args, **kwargs)

    def report_error(self) -> None:
        """Report an error."""
        self._metadata.report_error()

    def validate_environment(
        self,
        *,
        all_packages: bool = False,
        on_difference: Literal["warn", "error", "ignore", "return"] = "warn",
        exempt_packages: Optional[list[str]] = None,
    ) -> Union[bool, str]:
        """Validate the environment.

        Parameters
        ----------
        all_packages : bool, optional
            Check all packages in the environment (True) or just anemoi's (False), by default False.
        on_difference : Literal['warn', 'error', 'ignore', 'return'], optional
            What to do on difference, by default "warn"
        exempt_packages : list[str], optional
            List of packages to exempt from the check, by default EXEMPT_PACKAGES

        Returns
        -------
        Union[bool, str]
            boolean if `on_difference` is not 'return', otherwise formatted text of the differences
            True if environment is valid, False otherwise
        """
        return self._metadata.validate_environment(
            all_packages=all_packages, on_difference=on_difference, exempt_packages=exempt_packages
        )

    def open_dataset_args_kwargs(
        self, *, use_original_paths: bool, from_dataloader: Optional[Any] = None
    ) -> Tuple[Any, Any]:
        """Get arguments and keyword arguments for opening the dataset.

        Parameters
        ----------
        use_original_paths : bool
            Whether to use original paths.
        from_dataloader : Optional[Any], optional
            Data loader, by default None.

        Returns
        -------
        Tuple[Any, Any]
            Arguments and keyword arguments for opening the dataset.
        """
        return self._metadata.open_dataset_args_kwargs(
            use_original_paths=use_original_paths,
            from_dataloader=from_dataloader,
        )

    def constant_forcings_inputs(self, runner: Any, input_state: State) -> List[Forcings]:
        """Get constant forcings inputs.

        Parameters
        ----------
        runner : Any
            The runner.
        input_state : State
            The input state.

        Returns
        -------
        List[Forcings]
            The constant forcings inputs.
        """
        return self._metadata.constant_forcings_inputs(runner, input_state)

    def dynamic_forcings_inputs(self, runner: Any, input_state: State) -> List[Forcings]:
        """Get dynamic forcings inputs.

        Parameters
        ----------
        runner : Any
            The runner.
        input_state : State
            The input state.

        Returns
        -------
        List[Forcings]
            The dynamic forcings inputs.
        """
        return self._metadata.dynamic_forcings_inputs(runner, input_state)

    def boundary_forcings_inputs(self, runner: Any, input_state: State) -> List[Forcings]:
        """Get boundary forcings inputs.

        Parameters
        ----------
        runner : Any
            The runner.
        input_state : State
            The input state.

        Returns
        -------
        List[Forcings]
            The boundary forcings inputs.
        """
        return self._metadata.boundary_forcings_inputs(runner, input_state)

    def name_fields(self, fields: Any, namer: Optional[Callable[..., str]] = None) -> Any:
        """Name fields.

        Parameters
        ----------
        fields : Any
            The fields to name.
        namer : Optional[Callable[...,str]], optional
            The namer, by default None.

        Returns
        -------
        Any
            The named fields.
        """
        return self._metadata.name_fields(fields, namer=namer)

    def sort_by_name(
        self, fields: ekd.FieldList, *args: Any, namer: Optional[Callable[..., str]] = None, **kwargs: Any
    ) -> ekd.FieldList:
        """Sort fields by name.

        Parameters
        ----------
        fields : ekd.FieldList
            The fields to sort.
        *args : Any
            Additional arguments.
        namer : Optional[Callable[...,str]], optional
            The namer, by default None.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        ekd.FieldList
            The sorted fields.
        """
        return self._metadata.sort_by_name(fields, *args, namer=namer, **kwargs)

    def print_indices(self) -> None:
        """Print the indices."""
        return self._metadata.print_indices()

    def variable_categories(self) -> Any:
        """Get the variable categories.

        Returns
        -------
        Any
            The variable categories.
        """
        return self._metadata.variable_categories()

    def load_supporting_array(self, name: str) -> Any:
        """Load a supporting array.

        Parameters
        ----------
        name : str
            The name of the supporting array.

        Returns
        -------
        Any
            The supporting array.
        """
        return self._metadata.load_supporting_array(name)

    @property
    def supporting_arrays(self) -> Any:
        """Get the supporting arrays."""
        return self._metadata.supporting_arrays

    ###########################################################################

    @cached_property
    def lagged(self) -> List[datetime.timedelta]:
        """Return the list of steps for the `multi_step_input` fields."""
        result = list(range(0, self._metadata.multi_step_input))
        result = [-s * self._metadata.timestep for s in result]
        return sorted(result)

    @property
    def multi_step_input(self) -> Any:
        """Get the multi-step input."""
        return self._metadata.multi_step_input

    def print_variable_categories(self) -> None:
        """Print the variable categories."""
        return self._metadata.print_variable_categories()

    ###########################################################################
    # Data retrieval
    ###########################################################################

    def variables_from_input(self, *, include_forcings: bool) -> Any:
        """Get variables from input.

        Parameters
        ----------
        include_forcings : bool
            Whether to include forcings.

        Returns
        -------
        Any
            The variables from input.
        """
        return self._metadata.variables_from_input(include_forcings=include_forcings)

    @property
    def grid(self) -> Any:
        """Get the grid."""
        return self._metadata.grid

    @property
    def area(self) -> Any:
        """Get the area."""
        return self._metadata.area

    def mars_by_levtype(self, levtype: str) -> Any:
        """Get MARS requests by level type.

        Parameters
        ----------
        levtype : str
            The level type.

        Returns
        -------
        Any
            The MARS requests by level type.
        """
        return self._metadata.mars_by_levtype(levtype)

    def mars_requests(
        self,
        *,
        variables: List[str],
        dates: List[Date],
        use_grib_paramid: bool = False,
        always_split_time: bool = False,
        patch_request: Optional[Callable[[DataRequest], DataRequest]] = None,
        **kwargs: Any,
    ) -> List[DataRequest]:
        """Generate MARS requests for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            The list of variables.
        dates : List[Any]
            The list of dates.
        use_grib_paramid : bool, optional
            Whether to use GRIB paramid, by default False.
        always_split_time : bool, optional
            Whether to always split time, by default False.
        patch_request : Optional[Callable], optional
            A callable to patch the request, by default None.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        List[DataRequest]
            The list of MARS requests.
        """
        from anemoi.utils.grib import shortname_to_paramid
        from earthkit.data.utils.availability import Availability

        assert variables, "No variables provided"

        if not isinstance(dates, (list, tuple)):
            dates = [dates]

        dates = [to_datetime(d) for d in dates]

        assert dates, "No dates provided"

        result: List[DataRequest] = []

        DEFAULT_KEYS = ("class", "expver", "type", "stream", "levtype")
        DEFAULT_KEYS_AND_TIME = ("class", "expver", "type", "stream", "levtype", "time")

        # ECMWF operational data has stream oper for 00 and 12 UTC and scda for 06 and 18 UTC
        # The split oper/scda is a bit special
        KEYS = {("oper", "fc"): DEFAULT_KEYS_AND_TIME, ("scda", "fc"): DEFAULT_KEYS_AND_TIME}

        requests = defaultdict(list)

        for r in self._metadata.mars_requests(variables=variables):
            for date in dates:

                r = r.copy()

                base = date
                step = str(r.get("step", 0)).split("-")[-1]
                step = int(step)
                base = base - datetime.timedelta(hours=step)

                r["date"] = base.strftime("%Y-%m-%d")
                r["time"] = base.strftime("%H%M")

                r.update(kwargs)  # We do it here so that the Availability can use that information

                if always_split_time:
                    keys = DEFAULT_KEYS_AND_TIME
                else:
                    keys = KEYS.get((r.get("stream"), r.get("type")), DEFAULT_KEYS)
                key = tuple(r.get(k) for k in keys)

                # Special case because of oper/scda

                requests[key].append(r)

        result = []
        for reqs in requests.values():

            compressed = Availability(reqs)
            for r in compressed.iterate():

                if not r:
                    continue

                changed = True
                while changed:
                    changed = False
                    for k, v in r.items():
                        if isinstance(v, tuple):
                            r[k] = list(v)
                            changed = True

                        if isinstance(v, (list, tuple)) and len(v) == 1:
                            r[k] = v[0]
                            changed = True

                # Convert all to lists
                for k, v in r.items():
                    if not isinstance(v, list):
                        r[k] = [v]

                # Patch BEFORE the shortname to paramid

                if patch_request:
                    r = patch_request(r)

                # Convert all to lists (again)
                for k, v in r.items():
                    if isinstance(v, tuple):
                        r[k] = list(*v)
                    if not isinstance(v, list):
                        r[k] = [v]

                if use_grib_paramid and "param" in r:
                    r["param"] = [shortname_to_paramid(_) for _ in r["param"]]

                # Simplyfie the request

                for k, v in r.items():

                    if len(v) == 1:
                        r[k] = v[0]

                result.append(r)

        return result

    ###########################################################################
    # supporting arrays
    ###########################################################################

    @cached_property
    def _supporting_arrays(self) -> Any:
        """Get the supporting arrays."""
        return self._metadata._supporting_arrays

    @property
    def name(self) -> Any:
        """Get the name."""
        return self._metadata.name

    ###########################################################################
    # Misc
    ###########################################################################

    def provenance_training(self) -> Any:
        """Get the provenance of the training.

        Returns
        -------
        Any
            The provenance of the training.
        """
        return self._metadata.provenance_training()


class SourceCheckpoint(Checkpoint):
    """A checkpoint that represents a source."""

    def __init__(self, owner: Checkpoint, metadata: Any) -> None:
        """Initialize the SourceCheckpoint.

        Parameters
        ----------
        owner : Checkpoint
            The owner checkpoint.
        metadata : Any
            The metadata for the source checkpoint.
        """
        super().__init__(owner.path)
        self._owner = owner
        self._metadata = metadata

    def __repr__(self) -> str:
        """Represent the SourceCheckpoint as a string.

        Returns
        -------
        str
            String representation of the SourceCheckpoint.
        """
        return f"Source({self.name}@{self.path})"



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List

from anemoi.utils.logs import enable_logging_name

from anemoi.inference.task import Task
from anemoi.inference.types import State

LOG = logging.getLogger(__name__)


class Coupling:
    """Represents a coupling between a source and a target with specific variables."""

    def __init__(self, source: Task, target: Task, variables: List[str]) -> None:
        """Initialize a Coupling instance.

        Parameters
        ----------
        source : Task
            The source of the coupling.
        target : Task
            The target of the coupling.
        variables : List[str]
            The variables involved in the coupling.
        """
        self.source = source
        self.target = target
        self.variables: List[str] = variables

    def __repr__(self) -> str:
        """Return a string representation of the coupling.

        Returns
        -------
        str
            The string representation of the coupling.
        """
        return f"{self.source}->{self.target}(variables={self.variables})"


class CouplingSend(Coupling):
    """Represents a coupling send operation."""

    def apply(
        self,
        task: Task,
        transport: "Transport",
        *,
        input_state: State,
        output_state: State,
        constants: State,
        tag: int,
    ) -> None:
        """Apply the coupling send operation.

        Parameters
        ----------
        task : Task
            The task to apply the coupling to.
        transport : Transport
            The transport instance to use.
        input_state : State
            The input state dictionary.
        output_state : State
            The output state dictionary.
        constants : State
            The constants dictionary.
        tag : int
            The tag for the operation.
        """
        transport.send_state(
            task,
            self.target,
            input_state=input_state,
            variables=self.variables,
            constants=constants,
            tag=tag,
        )


class CouplingRecv(Coupling):
    """Represents a coupling receive operation."""

    def apply(
        self,
        task: Task,
        transport: "Transport",
        *,
        input_state: State,
        output_state: State,
        constants: State,
        tag: int,
    ) -> None:
        """Apply the coupling receive operation.

        Parameters
        ----------
        task : Any
            The task to apply the coupling to.
        transport : Transport
            The transport instance to use.
        input_state : State
            The input state dictionary.
        output_state : State
            The output state dictionary.
        constants : State
            The constants dictionary.
        tag : str
            The tag for the operation.
        """
        transport.receive_state(
            task,
            self.source,
            output_state=output_state,
            variables=self.variables,
            tag=tag,
        )


class Transport(ABC):
    """Abstract base class for transport mechanisms."""

    def __init__(self, couplings: List[Dict[str, List[str]]], tasks: Dict[str, Task]) -> None:
        """Initialize a Transport instance.

        Parameters
        ----------
        couplings : List[Dict[str,str]]
            The list of dict representing the couplings.
        tasks : Dict[str, Task]
            The dictionary of tasks.
        """
        enable_logging_name("main")
        self._couplings = couplings
        self.tasks: Dict[str, Task] = tasks

    @abstractmethod
    def send(self, sender: Task, target: Task, state: State, tag: int) -> None:
        """Send the state from the sender to the target.

        Parameters
        ----------
        sender : Task
            The sender of the state.
        target : Task
            The target
        state : State
            The state dictionary to send.
        tag : int
            The tag for the operation.
        """
        pass

    @abstractmethod
    def receive(self, receiver: Task, source: Task, tag: int) -> State:
        """Receive the state from the source to the receiver.

        Parameters
        ----------
        receiver : Task
            The receiver of the state.
        source : Task
            The source of the state.
        tag : int
            The tag for the operation.

        Returns
        -------
        State
            The received state dictionary.
        """
        pass

    def __repr__(self) -> str:
        """Return a string representation of the transport.

        Returns
        -------
        str
            The string representation of the transport.
        """
        return f"{self.__class__.__name__}()"

    def couplings(self, task: Task) -> List[Coupling]:
        """Get the couplings for a given task.

        Parameters
        ----------
        task : Task
            The task to get the couplings for.

        Returns
        -------
        List[Coupling]
            The list of couplings for the task.
        """
        couplings: List[Coupling] = []
        for coupling in self._couplings:
            assert isinstance(coupling, dict)
            assert len(coupling) == 1
            k, variables = list(coupling.items())[0]
            source, target = k.split("->")
            source = source.strip()
            target = target.strip()

            if task.name == source:
                couplings.append(
                    CouplingSend(
                        self.tasks[source],
                        self.tasks[target],
                        variables,
                    )
                )
            if task.name == target:
                couplings.append(
                    CouplingRecv(
                        self.tasks[source],
                        self.tasks[target],
                        variables,
                    )
                )

        return couplings

    def send_state(
        self,
        sender: Task,
        target: Task,
        *,
        input_state: State,
        variables: List[str],
        constants: Dict[str, Any],
        tag: int,
    ) -> None:
        """Send the state from the sender to the target.

        Parameters
        ----------
        sender : Any
            The sender of the state.
        target : Any
            The target of the state.
        input_state : Dict[str, Any]
            The input state dictionary.
        variables : List[str]
            The list of variables to send.
        constants : Dict[str, Any]
            The constants dictionary.
        tag : int
            The tag for the operation.
        """
        assert sender.name != target.name, f"Cannot send to self {sender}"

        fields: Dict[str, Any] = input_state["fields"]

        LOG.info(f"{sender}: sending to {target} {variables} {input_state['date']}")
        LOG.info("State fields: %s", list(fields.keys()))

        fields = {v: fields[v] for v in variables if v in fields}

        for v in variables:
            if v not in fields:
                # Check in the constants
                if v in constants:
                    LOG.warning(f"{sender}: {v} not in fields, using the value from constants")
                    fields[v] = constants[v]
                else:
                    raise ValueError(f"{sender}: Variable {v} not in fields or constants")

        for f, v in fields.items():
            assert len(v.shape) == 1, f"Expected  got {v.shape}"

        state: State = input_state.copy()
        state["fields"] = fields

        # Don't send unnecessary data
        state["latitudes"] = None
        state["longitudes"] = None
        for s in list(state.keys()):
            if s.startswith("_"):
                del state[s]

        self.send(sender, target, state, tag)

    def receive_state(
        self,
        receiver: Task,
        source: Task,
        *,
        output_state: State,
        variables: List[str],
        tag: int,
    ) -> None:
        """Receive the state from the source to the receiver.

        Parameters
        ----------
        receiver : Any
            The receiver of the state.
        source : Any
            The source of the state.
        output_state : Dict[str, Any]
            The output state dictionary.
        variables : List[str]
            The list of variables to receive.
        tag : int
            The tag for the operation.
        """
        assert receiver.name != source.name, f"Cannot receive from self {receiver}"

        state: State = self.receive(receiver, source, tag)

        assert isinstance(state, dict), state
        assert "fields" in state
        assert isinstance(state["fields"], dict), f"Expected dict got {type(state['fields'])}"

        output_state.setdefault("fields", {})

        fields_in: Dict[str, Any] = state["fields"]
        fields_out: Dict[str, Any] = output_state["fields"]

        for v in variables:
            if v in fields_out:
                raise ValueError(f"Variable {v} already in output state")

            if v not in fields_in:
                raise ValueError(f"Variable {v} not in input state")

            fields_out[v] = fields_in[v]

            # assert len(fields_out[v].shape) == 1, f"Expected  got {fields_out[v].shape}"



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import earthkit.data as ekd
import yaml
from anemoi.utils.registry import Registry

from anemoi.inference.config import Configuration

LOG = logging.getLogger(__name__)


template_provider_registry = Registry(__name__)


def create_template_provider(owner: Any, config: Configuration) -> "TemplateProvider":
    """Create a template provider from the given configuration.

    Parameters
    ----------
    owner : Any
        The owner of the template provider.
    config : Configuration
        The configuration for the template provider.

    Returns
    -------
    TemplateProvider
        The created template provider.
    """
    return template_provider_registry.from_config(config, owner)


class TemplateProvider:
    """Base class for template providers."""

    def __init__(self, manager: Any) -> None:
        """Initialize the template provider.

        Parameters
        ----------
        manager : Any
            The manager for the template provider.
        """
        self.manager = manager

    def template(self, variable: str, lookup: Dict[str, Any]) -> ekd.Field:
        """Get the template for the given variable and lookup.

        Parameters
        ----------
        variable : str
            The variable to get the template for.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        ekd.Field
            The template field.
        """
        raise NotImplementedError()


class IndexTemplateProvider(TemplateProvider):
    """Template provider based on an index file."""

    def __init__(self, manager: Any, index_path: str) -> None:
        """Initialize the index template provider.

        Parameters
        ----------
        manager : Any
            The manager for the template provider.
        index_path : str
            The path to the index file.
        """
        super().__init__(manager)
        self.index_path = index_path

        with open(index_path) as f:
            self.templates = yaml.safe_load(f)

        if not isinstance(self.templates, list):
            raise ValueError("Invalid templates.yaml, must be a list")

        # TODO: use pydantic
        for template in self.templates:
            if not isinstance(template, list):
                raise ValueError(f"Invalid template in templates.yaml, must be a list: {template}")
            if len(template) != 2:
                raise ValueError(f"Invalid template in templates.yaml, must have exactly 2 elements: {template}")

            match, grib = template
            if not isinstance(match, dict):
                raise ValueError(f"Invalid match in templates.yaml, must be a dict: {match}")

            if not isinstance(grib, str):
                raise ValueError(f"Invalid grib in templates.yaml, must be a string: {grib}")

    def template(self, variable: str, lookup: Dict[str, Any]) -> Optional[ekd.Field]:
        """Get the template for the given variable and lookup.

        Parameters
        ----------
        variable : str
            The variable to get the template for.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        Optional[ekd.Field]
            The template field if found, otherwise None.
        """

        def _as_list(value: Any) -> List[Any]:
            if not isinstance(value, list):
                return [value]
            return value

        for template in self.templates:
            match, grib = template
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("%s", [(lookup.get(k), _as_list(v)) for k, v in match.items()])

            if all(lookup.get(k) in _as_list(v) for k, v in match.items()):
                return self.load_template(grib, lookup)

        return None

    def load_template(self, grib: str, lookup: Dict[str, Any]) -> Optional[ekd.Field]:
        """Load the template for the given GRIB and lookup.

        Parameters
        ----------
        grib : str
            The GRIB string.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        Optional[ekd.Field]
            The loaded template field if found, otherwise None.
        """
        raise NotImplementedError()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import List
from typing import Optional

import earthkit.data as ekd

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import State

from ..decorators import main_argument
from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


@input_registry.register("grib")
@main_argument("path")
class GribFileInput(GribInput):
    """Handles grib files."""

    trace_name = "grib file"

    def __init__(self, context: Context, path: str, *, namer: Optional[Any] = None, **kwargs: Any) -> None:
        """Initialize the GribFileInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        path : str
            The path to the GRIB file.
        namer : Optional[Any]
            Optional namer for the input.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, namer=namer, **kwargs)
        self.path = path

    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """
        return self._create_input_state(ekd.from_source("file", self.path), variables=None, date=date)

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variables to load.
        dates : List[Date]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        State
            The loaded forcings state.
        """
        return self._load_forcings_state(
            ekd.from_source("file", self.path),
            variables=variables,
            dates=dates,
            current_state=current_state,
        )



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
from typing import Any
from typing import Dict

import torch
from anemoi.utils.config import DotDict

from anemoi.inference.checkpoint import Checkpoint
from anemoi.inference.config import LOG
from anemoi.inference.metadata import Metadata
from anemoi.inference.testing import float_hash


class MockModel(torch.nn.Module):
    """Mock model for testing."""

    def __init__(self, medatada: Dict[str, Any], supporting_arrays: Dict[str, Any]) -> None:
        """Initialize the mock model.

        Parameters
        ----------
        medatada : dict
            The metadata for the model.
        supporting_arrays : dict
            The supporting arrays for the model.
        """
        super().__init__()
        metadata = DotDict(medatada)

        self.features_in = len(metadata.data_indices.model.input.full)
        self.features_out = len(metadata.data_indices.model.output.full)
        self.roll_window = metadata.config.training.multistep_input
        self.grid_size = metadata.dataset.shape[-1]

        self.input_shape = (1, self.roll_window, self.grid_size, self.features_in)
        self.output_shape = (1, 1, self.grid_size, self.features_out)

        checkpoint = Checkpoint(Metadata(metadata))
        self.input_variables = {v: k for k, v in checkpoint.variable_to_input_tensor_index.items()}
        self.output_variables = checkpoint.output_tensor_index_to_variable
        self.lagged = checkpoint.lagged

        self.typed_variables = checkpoint.typed_variables
        self.timestep = checkpoint.timestep

        self.first = True
        self.constant_in_time: Dict[int, torch.Tensor] = {}

    def _check(
        self,
        x: torch.Tensor,
        lag: int,
        feature: int,
        date: datetime.datetime,
        step: datetime.timedelta,
        tensor_dates: list[datetime.datetime],
        value: float,
        expect: torch.Tensor,
        title: str,
    ) -> None:
        """Check if the tensor values are as expected.

        Parameters
        ----------
        x : torch.Tensor
            The input tensor.
        lag : int
            The lag index.
        feature : int
            The feature index.
        date : datetime.datetime
            The current date.
        step : datetime.timedelta
            The time step.
        tensor_dates : list of datetime.datetime
            The list of tensor dates.
        value : float
            The expected value.
        expect : torch.Tensor
            The expected tensor.
        title : str
            The title of the check.
        """
        if not torch.isclose(x[:, lag, :, feature], expect).all():
            LOG.error(
                "Feature: %s (%s), lag: %s (%s)",
                feature,
                self.input_variables[feature],
                lag,
                tensor_dates[lag],
            )
            LOG.error("Date: %s, Step: %s, Tensor dates: %s", date, step, tensor_dates)
            LOG.error(
                "%s: expected %s for %s at lag %s, got %s",
                title,
                value,
                self.input_variables[feature],
                tensor_dates[lag],
                x[:, lag, :, feature][..., 0],
            )
            LOG.error("x: %s", x[:, :, :, feature])

            raise ValueError(
                f"{title}: expected {expect} for {self.input_variables[feature]} at lag {lag}, got {x[:, lag, :, feature][...,0]}"
            )

    def predict_step(
        self, x: torch.Tensor, date: datetime.datetime, step: datetime.timedelta, **kwargs: Any
    ) -> torch.Tensor:
        """Perform a prediction step.

        Parameters
        ----------
        x : torch.Tensor
            The input tensor.
        date : datetime.datetime
            The current date.
        step : datetime.timedelta
            The time step.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            The output tensor.
        """
        assert x.shape == self.input_shape, f"Expected {self.input_shape}, got {x.shape}"

        # Date of the data
        tensor_dates = [(date - self.timestep) + h for h in self.lagged]

        for lag in range(x.shape[1]):
            for feature in range(x.shape[3]):
                value = float_hash(self.input_variables[feature], tensor_dates[lag])
                expect = torch.Tensor([value])
                variable = self.typed_variables[self.input_variables[feature]]

                if not variable.is_computed_forcing and not variable.is_constant_in_time:
                    # Normal prognostice variable
                    self._check(x, lag, feature, date, step, tensor_dates, value, expect, "Prognostics")
                    continue

                if variable.is_constant_in_time:
                    # Constant in from input, like lsm, only check the first time
                    if self.first:
                        if variable.is_computed_forcing:
                            # Computed forcing, such as cos_latitude. We trust the value to be correct
                            pass
                        else:
                            # Input constant, such as LSM
                            self._check(x, lag, feature, date, step, tensor_dates, value, expect, "Constants")
                        self.constant_in_time[feature] = x[:, lag, :, feature]
                    else:
                        # Check that the value is the same as the first time
                        assert torch.isclose(x[:, lag, :, feature], self.constant_in_time[feature]).all()
                    continue

                if variable.is_computed_forcing:
                    # Computed forcing, such as insolation. We trust the value to be correct
                    continue

                assert False, f"Unknown variable type {variable}"

            self.first = False

        y = torch.zeros(self.output_shape)

        for feature in range(y.shape[3]):
            value = float_hash(self.output_variables[feature], date)
            y[:, :, :, feature] = torch.ones(y[:, :, :, feature].shape) * value

        return y



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import warnings
from collections import defaultdict
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict

from anemoi.inference.types import DataRequest

from .protocol import MetadataProtocol

LOG = logging.getLogger(__name__)


def warn(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to issue a warning when using legacy functions.

    Parameters
    ----------
    func : function
        The legacy function to be wrapped.

    Returns
    -------
    function
        The wrapped function with a warning.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        warnings.warn(f"Using legacy {func.__name__}, please try to patch your weights.")
        return func(*args, **kwargs)

    return wrapper


class LegacyMixin(MetadataProtocol):

    # `self` is a `Metadata` object

    @warn
    def _legacy_variables_metadata(self) -> Dict[str, Dict[str, Any]]:
        """Generate metadata for legacy variables.

        Returns
        -------
        Dict[str, Dict[str, Any]]
            Metadata for each variable.
        """
        result: Dict[str, Any] = {}
        unkowns = []
        for variable in self.variables:
            if variable in (
                "insolation",
                "cos_solar_zenith_angle",
                "cos_julian_day",
                "sin_julian_day",
                "cos_local_time",
                "sin_local_time",
            ):
                result[variable] = dict(computed_forcing=True, constant_in_time=False)
                continue

            if variable in ("cos_latitude", "cos_longitude", "sin_latitude", "sin_longitude"):
                result[variable] = dict(computed_forcing=True, constant_in_time=True)
                continue

            param_level = variable.split("_")
            if len(param_level) == 2:
                # Assumes levtype=pl for now
                levtype = "pl"

                try:
                    levelist = int(param_level[1])
                    mars = dict(param=param_level[0], levelist=levelist, levtype=levtype)
                    result[variable] = dict(mars=mars)
                    continue
                except ValueError:
                    pass

            if variable in (
                "z",
                "lsm",
                "sdor",
                "slor",
                "cl",
                "cvh",
                "cvl",
                "slt",
                "tvh",
                "tvl",
            ):
                mars = dict(param=variable, levtype="sfc")
                result[variable] = dict(mars=mars, constant_in_time=True)
                continue

            if variable in ("cp", "tp", "sf", "ro", "ssrd", "strd"):
                mars = dict(param=variable, levtype="sfc", type="fc", stream="oper")
                result[variable] = dict(mars=mars, process="accumulation", period=(0, 6))

                continue

            if variable in ("sp", "msl", "10u", "10v", "2t", "2d", "skt", "tcw"):
                mars = dict(param=variable, levtype="sfc")
                result[variable] = dict(mars=mars)
                continue

            unkowns.append(variable)
            mars = dict(param=variable, levtype="sfc")
            result[variable] = dict(mars=mars)

        if unkowns:
            warnings.warn(f"Unknown variables: {unkowns}, assuming an/sfc")

        return result

    def _legacy_check_variables_metadata(self, variables: Dict[str, Dict[str, Any]]) -> None:
        """Check and update metadata for legacy variables.

        Parameters
        ----------
        variables : Dict[str, Dict[str, Any]]
            The metadata dictionary to be checked and updated.
        """
        if variables == {}:
            variables.update(self._legacy_variables_metadata())
            return

        first = True
        for variable, metadata in variables.items():
            if "mars" not in metadata:
                if "param" in metadata:
                    if first:
                        LOG.warning("Old metadata format detected. Please update your weights.")
                        first = False
                    metadata["mars"] = metadata.copy()

            if variable in (
                "insolation",
                "cos_solar_zenith_angle",
                "cos_julian_day",
                "sin_julian_day",
                "cos_local_time",
                "sin_local_time",
            ):
                ok = metadata.get("computed_forcing") is True and metadata.get("constant_in_time") is False
                if not ok:
                    if first:
                        LOG.warning("Old metadata format detected. Please update your weights.")
                        first = False
                    metadata["computed_forcing"] = True
                    metadata["constant_in_time"] = False

            if variable in ("cos_latitude", "cos_longitude", "sin_latitude", "sin_longitude"):
                ok = metadata.get("computed_forcing") is True and metadata.get("constant_in_time") is True
                if not ok:
                    if first:
                        LOG.warning("Old metadata format detected. Please update your weights.")
                        first = False
                    metadata["computed_forcing"] = True
                    metadata["constant_in_time"] = True

    @warn
    def _legacy_number_of_grid_points(self) -> int:
        """Get the number of grid points for the legacy grid.

        Returns
        -------
        int
            Number of grid points.
        """
        POINTS = {"o96": 40_320, "n320": 542_080}

        return POINTS[self.grid.lower()]

    @warn
    def _legacy_data_request(self) -> DataRequest:
        """Retrieve the data request from metadata.

        Returns
        -------
        DataRequest
            The data request information.

        Raises
        ------
        ValueError
            If no data request is found in metadata.
        """
        from anemoi.utils.config import find

        result = find(self._metadata["dataset"], "data_request")
        if len(result) == 0:
            raise ValueError("No data_request found in metadata")

        if len(result) > 1:
            check = ("grid", "area")
            checks = defaultdict(set)
            for r in result:
                for c in check:
                    checks[c].add(str(r.get(c)))

            for c in check:
                if len(checks[c]) > 1:
                    warnings.warn(f"{c} is ambigous: {checks[c]}")

            result = [r for r in result if r["grid"] is not None]

        return result[0]



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import earthkit.data as ekd

LOG = logging.getLogger(__name__)


GRIB1_ONLY: List[str] = []

GRIB2_ONLY: List[str] = ["typeOfGeneratingProcess"]


ORDERING = (
    "edition",
    "typeOfLevel",
    "stepType",
    "productDefinitionTemplateNumber",
    "eps",
    "number",
)

ORDERING = {k: i for i, k in enumerate(ORDERING)}


def _ordering(item: tuple) -> int:
    """Get the ordering index for a given item.

    Parameters
    ----------
    item : tuple
        The item to get the ordering index for.

    Returns
    -------
    int
        The ordering index.
    """
    return ORDERING.get(item[0], 999)


def _param(param: Any) -> str:
    """Determine the parameter type based on its value.

    Parameters
    ----------
    param : str
        The parameter value.

    Returns
    -------
    str
        The parameter type.
    """
    try:
        int(param)
        return "paramId"
    except ValueError:
        try:
            float(param)
            return "param"
        except ValueError:
            return "shortName"


def _step_in_hours(step: Any) -> int:
    """Convert a step to hours.

    Parameters
    ----------
    step : Any
        The step to convert.

    Returns
    -------
    int
        The step in hours.
    """
    step = step.total_seconds() / 3600
    assert int(step) == step
    return int(step)


STEP_TYPE = {
    "accumulation": "accum",
    "average": "avg",
    "maximum": "max",
    "minimum": "min",
    "instantaneous": None,
}


def encode_time_processing(
    *,
    result: Dict[str, Any],
    template: ekd.Field,
    variable: Any,
    step: Any,
    previous_step: Optional[Any],
    start_steps: Dict[Any, Any],
    edition: int,
    ensemble: bool,
) -> None:
    """Encode time processing information into the result dictionary.

    Parameters
    ----------
    result : Dict[str, Any]
        The result dictionary to update.
    template : ekd.Field
        The template field.
    variable : Any
        The variable containing time processing information.
    step : Any
        The current step.
    previous_step : Optional[Any]
        The previous step.
    start_steps : Dict[Any, Any]
        The start steps dictionary.
    edition : int
        The GRIB edition.
    ensemble : bool
        Whether the data is part of an ensemble.
    """
    assert edition in (1, 2)

    if variable.time_processing is None:
        result["step"] = _step_in_hours(step)
        # result["startStep"] = _step_in_hours(step)
        # result["endStep"] = _step_in_hours(step)
        result["stepType"] = "instant"
        return

    if previous_step is None:
        if not variable.is_accumulation:
            LOG.warning(f"No previous step available for time processing `{variable.time_processing}` for `{variable}`")
        previous_step = step

    start = _step_in_hours(start_steps.get(variable, previous_step))
    end = _step_in_hours(step)

    result["startStep"] = start
    result["endStep"] = end
    result["stepType"] = STEP_TYPE[variable.time_processing]

    if edition == 1:
        return

    if ensemble:
        result["productDefinitionTemplateNumber"] = 11
    else:
        result["productDefinitionTemplateNumber"] = 8


LEVTYPES = {
    "pl": "isobaricInhPa",
    "ml": "hybrid",
    "pt": "theta",
    "pv": "potentialVorticity",
}


def grib_keys(
    *,
    values: Any,
    template: Any,
    variable: Any,
    ensemble: bool,
    param: Optional[Union[int, float, str]],
    date: int,
    time: int,
    step: Any,
    previous_step: Optional[Any],
    start_steps: Dict[Any, Any],
    keys: Dict[str, Any],
    grib1_keys: Dict[Union[int, float, str], Dict[str, Any]] = {},
    grib2_keys: Dict[Union[int, float, str], Dict[str, Any]] = {},
) -> Dict[str, Any]:
    """Generate GRIB keys for encoding.

    Parameters
    ----------
    values : Any
        The values to encode.
    template : Any
        The template to use.
    variable : Any
        The variable containing GRIB keys.
    ensemble : bool
        Whether the data is part of an ensemble.
    param : Optional[Union[int, float, str]]
        The parameter value.
    date : int
        The date value.
    time : int
        The time value.
    step : Any
        The current step.
    previous_step : Optional[Any]
        The previous step.
    start_steps : Dict[Any, Any]
        The start steps dictionary.
    keys : Dict[str, Any]
        The initial keys dictionary.
    grib1_keys : Dict[Union[int, float, str], Dict[str, Any]], optional
        Additional GRIB1 keys.
    grib2_keys : Dict[Union[int, float, str], Dict[str, Any]], optional
        Additional GRIB2 keys.

    Returns
    -------
    Dict[str, Any]
        The generated GRIB keys.
    """
    result = keys.copy()

    edition = keys.get("edition")
    if edition is None and template is not None:
        edition = template.metadata("edition")

    if edition is None:
        edition = 1

    if edition == 1:
        for k in GRIB2_ONLY:
            result.pop(k, None)

    if edition == 2:
        for k in GRIB1_ONLY:
            result.pop(k, None)

    result["edition"] = edition

    result["eps"] = 1 if ensemble else 0

    if param is not None:
        result.setdefault(_param(param), param)

        if edition == 1:
            result.update(grib1_keys.get(param, {}))

        if edition == 2:
            result.update(grib2_keys.get(param, {}))

    result.setdefault("type", "fc")

    if result.get("type") in ("an", "fc"):
        # For organisations that do not use type
        result.setdefault("dataType", result.pop("type"))

    result["date"] = date
    result["time"] = time

    encode_time_processing(
        result=result,
        template=template,
        variable=variable,
        step=step,
        previous_step=previous_step,
        start_steps=start_steps,
        edition=edition,
        ensemble=ensemble,
    )

    for k, v in variable.grib_keys.items():
        if k not in ("domain", "type", "stream", "expver", "class", "param", "number", "step", "date", "hdate", "time"):
            if k == "levtype":
                v = LEVTYPES.get(v)
                if v is None:
                    continue
                k = "typeOfLevel"
            result.setdefault(k, v)

    result = {k: v for k, v in sorted(result.items(), key=_ordering) if v is not None}

    return result


def check_encoding(handle: Any, keys: Dict[str, Any], first: bool = True) -> None:
    """Check if the GRIB encoding matches the expected keys.

    Parameters
    ----------
    handle : Any
        The GRIB handle.
    keys : Dict[str, Any]
        The expected keys.
    first : bool, optional
        Whether this is the first check.

    Raises
    ------
    ValueError
        If the GRIB field could not be encoded.
    """

    def same(w: Any, v: Any, k: str) -> bool:
        if type(v) is type(w):
            # Keep `mypy` happy
            equal: bool = v == w
            return equal
        return str(w) == str(v)

    mismatches = {}
    for k, v in keys.items():
        if k == "param":
            try:
                int(v)
                k = "paramId"
            except ValueError:
                try:
                    float(v)
                    k = "param"
                except ValueError:
                    k = "shortName"

        if k == "date":
            v = int(str(v).replace("-", ""))

        if k == "time":
            v = int(v)
            if v < 100:
                v *= 100

        if isinstance(v, int):
            w = handle.get_long(k)
        else:
            w = handle.get(k)

        if not same(w, v, k):
            mismatches[k] = 'Expected "{}" but got "{}"'.format(v, w)

    if mismatches:

        if first:
            import eccodes
            from earthkit.data.readers.grib.codes import GribCodesHandle

            handle = GribCodesHandle(eccodes.codes_clone(handle._handle), None, None)
            return check_encoding(handle, keys, first=False)

        raise ValueError(f"GRIB field could not be encoded. Mismatches={mismatches}")


def encode_message(
    *,
    values: Optional[Any],
    template: Any,
    metadata: Dict[str, Any],
    check_nans: bool = False,
    missing_value: Union[int, float] = 9999,
) -> Any:
    """Encode a GRIB message.

    Parameters
    ----------
    values : Optional[Any]
        The values to encode.
    template : Any
        The template to use.
    metadata : Dict[str, Any]
        The metadata for the GRIB message.
    check_nans : bool, optional
        Whether to check for NaNs in the values.
    missing_value : Union[int, float], optional
        The value to use for missing data.

    Returns
    -------
    Any
        The encoded GRIB handle.
    """
    metadata = metadata.copy()  # avoid modifying the original metadata
    handle = template.handle.clone()

    if check_nans and values is not None:
        import numpy as np

        if np.isnan(values).any():
            # missing_value = np.finfo(values.dtype).max
            missing_value = missing_value
            values = np.nan_to_num(values, nan=missing_value)
            metadata["missingValue"] = missing_value
            metadata["bitmapPresent"] = 1

    if int(metadata.get("deleteLocalDefinition", 0)):
        for k in ("class", "type", "stream", "expver", "setLocalDefinition"):
            metadata.pop(k, None)

    metadata.setdefault("generatingProcessIdentifier", 255)

    LOG.debug("GribOutput.metadata %s", metadata)

    single = {}
    multiple = {}
    for k, v in metadata.items():
        if isinstance(v, (int, float, str, bool)):
            single[k] = v
        else:
            multiple[k] = v

    try:
        # Try to set all metadata at once
        # This is needed when we set multiple keys that are interdependent
        handle.set_multiple(single)
    except Exception as e:
        LOG.error("Failed to set metadata at once: %s", e)
        # Try again, but one by one
        for k, v in single.items():
            handle.set(k, v)

    for k, v in multiple.items():
        handle.set(k, v)

    if values is not None:
        handle.set_values(values)

    return handle


class GribWriter:
    """Write GRIB messages to one or more files."""

    def __init__(self, path: str, split_output: bool = False) -> None:
        """Initialize the GribWriter.

        Parameters
        ----------
        path : str
            The path to the output file.
        split_output : bool, optional
            Whether to split the output into multiple files.
        """
        self._files: Dict[str, Any] = {}
        self.filename = path
        self.split_output: Optional[List[str]] = None

        if split_output:
            self.split_output = re.findall(r"\{(.*?)\}", self.filename)

    def close(self) -> None:
        """Close all open files."""
        for f in self._files.values():
            f.close()

    def __enter__(self) -> "GribWriter":
        """Enter the runtime context related to this object.

        Returns
        -------
        GribWriter
            The GribWriter instance.
        """
        return self

    def __exit__(self, exc_type: Optional[type], exc_value: Optional[BaseException], trace: Optional[Any]) -> None:
        """Exit the runtime context related to this object.

        Parameters
        ----------
        exc_type : Optional[type]
            The exception type.
        exc_value : Optional[BaseException]
            The exception value.
        trace : Optional[Any]
            The traceback object.
        """
        self.close()

    def write(
        self,
        *,
        values: Optional[Any],
        template: Any,
        metadata: Dict[str, Any],
        check_nans: bool = False,
        missing_value: Union[int, float] = 9999,
    ) -> tuple:
        """Write a GRIB message to the target file.

        Parameters
        ----------
        values : Optional[Any]
            The values to encode.
        template : Any
            The template to use.
        metadata : Dict[str, Any]
            The metadata for the GRIB message.
        check_nans : bool, optional
            Whether to check for NaNs in the values.
        missing_value : Union[int, float], optional
            The value to use for missing data.

        Returns
        -------
        tuple
            The encoded GRIB handle and the file path.
        """
        handle = encode_message(
            values=values,
            check_nans=check_nans,
            metadata=metadata,
            template=template,
            missing_value=missing_value,
        )

        file, path = self.target(handle)
        handle.write(file)

        return handle, path

    def target(self, handle: Any) -> tuple:
        """Determine the target file for the GRIB message.

        Parameters
        ----------
        handle : Any
            The GRIB handle.

        Returns
        -------
        tuple
            The file object and the file path.
        """
        if self.split_output:
            path = self.filename.format(**{k: handle.get(k) for k in self.split_output})
        else:
            path = self.filename

        if path not in self._files:
            self._files[path] = open(path, "wb")

        return self._files[path], path



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
import pickle
import struct
from typing import Any
from typing import Dict
from typing import Tuple

from anemoi.utils.logs import enable_logging_name
from anemoi.utils.logs import set_logging_name

from anemoi.inference.task import Task
from anemoi.inference.types import State

from ..transport import Transport
from . import transport_registry

LOG = logging.getLogger(__name__)


@transport_registry.register("processes")
class ProcessesTransport(Transport):
    """Transport implementation using processes."""

    def __init__(self, couplings: Any, tasks: Dict[str, Any], *args: Any, **kwargs: Any) -> None:
        """Initialize the ProcessesTransport.

        Parameters
        ----------
        couplings : Any
            The couplings for the transport.
        tasks : Dict[str, Any]
            The tasks to be executed.
        """
        super().__init__(couplings, tasks)
        self.children: Dict[int, str] = {}
        self.pipes: Dict[Tuple[str, str], Tuple[int, int]] = {}
        enable_logging_name("main")

    def child_process(self, task: Any) -> int:
        """Run the task in a child process.

        Parameters
        ----------
        task : Any
            The task to be run.

        Returns
        -------
        int
            The exit status of the child process.
        """
        set_logging_name(task.name)

        # Close all the pipes that are not needed
        for (task1, task2), (read_fd, write_fd) in self.pipes.items():
            if task.name not in (task1, task2):
                os.close(read_fd)
                os.close(write_fd)

        try:
            task.run(self)
        except Exception as e:
            LOG.exception(e)
            return 1
        return 0

    def start(self) -> None:
        """Start the transport by forking processes for each task."""

        # Many to many pipes. May not scale well
        # We can use the couplings to reduce the number of pipes
        self.pipes = {}
        for task1 in self.tasks:
            for task2 in self.tasks:
                if task1 != task2:
                    read_fd, write_fd = os.pipe()
                    os.set_inheritable(read_fd, True)
                    os.set_inheritable(write_fd, True)
                    self.pipes[(task1, task2)] = (read_fd, write_fd)

        for name, task in self.tasks.items():
            pid = os.fork()
            if pid == 0:
                os._exit(self.child_process(task))
            else:
                self.children[pid] = name

        # We need to close the pipes in the parent process
        for read_fd, write_fd in self.pipes.values():
            os.close(read_fd)
            os.close(write_fd)

    def wait(self) -> None:
        """Wait for all child processes to complete and handle any errors."""
        while self.children:
            (pid, status) = os.wait()
            LOG.info(f"Child process {pid} ({self.children[pid]}) exited with status {status}")
            del self.children[pid]

            if status != 0:
                for pid in self.children:
                    os.kill(pid, 15)

    def send(self, sender: Task, target: Task, state: State, tag: int) -> None:
        """Send a state from the sender to the target.

        Parameters
        ----------
        sender : Any
            The task sending the state.
        target : Any
            The task receiving the state.
        state : Any
            The state to be sent.
        tag : int
            The tag associated with the state.
        """
        # TODO: something more efficient than pickle
        _, write_fd = self.pipes[(sender.name, target.name)]
        pickle_data = pickle.dumps(state)

        os.write(write_fd, struct.pack("!Q", tag))
        os.write(write_fd, struct.pack("!Q", len(pickle_data)))
        os.write(write_fd, pickle_data)

    def receive(self, receiver: Task, source: Task, tag: int) -> State:
        """Receive a state from the source to the receiver.

        Parameters
        ----------
        receiver : Any
            The task receiving the state.
        source : Any
            The task sending the state.
        tag : int
            The tag associated with the state.

        Returns
        -------
        Any
            The received state.
        """
        read_fd, _ = self.pipes[(source.name, receiver.name)]

        recieved_tag = struct.unpack("!Q", os.read(read_fd, 8))[0]
        assert recieved_tag == tag, (recieved_tag, tag)

        size = struct.unpack("!Q", os.read(read_fd, 8))[0]
        data = os.read(read_fd, size)
        state = pickle.loads(data)
        if isinstance(state, Exception):
            raise state

        return state



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any


class FakeMetadata:
    """A class to simulate metadata for testing purposes.

    This class returns None for any attribute accessed, simulating the absence
    of metadata.
    """

    def __getattr__(self, name: str) -> Any:
        """Simulate the absence of metadata by returning None for any attribute.

        Parameters
        ----------
        name : str
            The name of the attribute being accessed.

        Returns
        -------
        Any
            Always returns None.
        """
        return None



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List

import earthkit.data as ekd
import numpy as np
from anemoi.transform.grids.unstructured import UnstructuredGridFieldList
from earthkit.data.indexing.fieldlist import FieldArray

from anemoi.inference.context import Context
from anemoi.inference.inputs.dataset import DatasetInput
from anemoi.inference.types import Date
from anemoi.inference.types import FloatArray
from anemoi.inference.types import IntArray
from anemoi.inference.types import State

LOG = logging.getLogger(__name__)


class Forcings(ABC):
    """Represents the forcings for the model."""

    def __init__(self, context: Context):
        """Initialize the Forcings object.

        Parameters
        ----------
        context : Context
            The context for the forcings.
        """
        self.context = context
        self.checkpoint = context.checkpoint
        self.kinds = dict(unknown=True)  # Used for debugging

    @abstractmethod
    def load_forcings_array(self, dates: List[Date], current_state: State) -> FloatArray:
        """Load the forcings for the given dates.

        Parameters
        ----------
        dates : List[Date]
            The list of dates for which to load the forcings.
        current_state : State
            The current state of the model.

        Returns
        -------
        FloatArray
            The loaded forcings as a numpy array.
        """
        pass

    def __repr__(self) -> str:
        """Return a string representation of the Forcings object."""
        return f"{self.__class__.__name__}"

    def _state_to_numpy(self, state: State, variables: List[str], dates: List[Date]) -> FloatArray:
        """Convert the state dictionary to a numpy array.
        This assumes that the state dictionary contains the fields for the given variables.
        And that the fields values are sorted by dates.

        Parameters
        ----------
        state : State
            The state dictionary.
        variables : List[str]
            The list of variables.
        dates : List[Date]
            The list of dates.

        Returns
        -------
        FloatArray
            The state as a numpy array.
        """
        fields = state["fields"]

        if len(dates) == 1:
            result = np.stack([fields[v] for v in variables], axis=0)
            if len(result.shape) == 2:
                result = result[:, np.newaxis]
        else:
            result = np.stack([fields[v] for v in variables], axis=0)

        # Assert that the shape is correct: (variables, dates, values)

        assert len(result.shape) == 3 and result.shape[0] == len(variables) and result.shape[1] == len(dates), (
            result.shape,
            variables,
            dates,
        )

        return result


class ComputedForcings(Forcings):
    """Compute forcings like `cos_julian_day` or `insolation`."""

    trace_name = "computed"

    def __init__(self, context: Context, variables: List[str], mask: Any):
        """Initialize the ComputedForcings object.

        Parameters
        ----------
        context : Context
            The context for the forcings.
        variables : List[str]
            The list of variables to compute.
        mask : Any
            The mask to apply to the forcings.
        """
        super().__init__(context)
        self.variables = variables
        self.mask = mask
        self.kinds = dict(computed=True)  # Used for debugging

    def __repr__(self) -> str:
        """Return a string representation of the ComputedForcings object."""
        return f"{self.__class__.__name__}({self.variables})"

    def load_forcings_array(self, dates: List[Date], current_state: State) -> FloatArray:
        """Load the computed forcings for the given dates.

        Parameters
        ----------
        dates : List[Date]
            The list of dates for which to load the forcings.
        current_state : State
            The current state of the model.

        Returns
        -------
        FloatArray
            The loaded forcings as a numpy array.
        """
        LOG.debug("Adding dynamic forcings %s", self.variables)

        if not isinstance(dates, (list, tuple)):
            dates = [dates]

        source = UnstructuredGridFieldList.from_values(
            latitudes=current_state["latitudes"],
            longitudes=current_state["longitudes"],
        )

        ds = ekd.from_source("forcings", source, date=dates, param=self.variables)

        assert len(ds) == len(self.variables) * len(dates), (len(ds), len(self.variables), dates)

        def rename(field: ekd.Field, _: str, metadata: Dict[str, Any]) -> str:
            return metadata["param"]

        ds = FieldArray([f.clone(name=rename) for f in ds])

        forcing = ds.order_by(name=self.variables, valid_datetime="ascending")

        # Forcing are sorted by `compute_forcings`  in the order (variable, date)

        return forcing.to_numpy(dtype=np.float32, flatten=True).reshape(len(self.variables), len(dates), -1)


class CoupledForcings(Forcings):
    """Retrieve forcings from the input."""

    @property
    def trace_name(self) -> str:
        """Return the trace name of the input."""
        return self.input.trace_name

    def __init__(self, context: Context, input: Any, variables: List[str], mask: IntArray):
        """Initialize the CoupledForcings object.

        Parameters
        ----------
        context : Context
            The context for the forcings.
        input : Any
            The input object.
        variables : List[str]
            The list of variables to retrieve.
        mask : IntArray
            The mask to apply to the forcings.
        """
        super().__init__(context)
        self.variables = variables
        self.mask = mask
        self.input = input
        self.kinds = dict(retrieved=True)  # Used for debugging

    def __repr__(self) -> str:
        """Return a string representation of the CoupledForcings object."""
        return f"{self.__class__.__name__}({self.variables})"

    def load_forcings_array(self, dates: List[Date], current_state: State) -> FloatArray:
        """Load the forcings for the given dates.

        Parameters
        ----------
        dates : List[Any]
            The list of dates for which to load the forcings.
        current_state : State
            The current state of the model.

        Returns
        -------
        FloatArray
            The loaded forcings as a numpy array.
        """
        return self._state_to_numpy(
            self.input.load_forcings_state(
                variables=self.variables,
                dates=dates,
                current_state=current_state,
            ),
            self.variables,
            dates,
        )


class BoundaryForcings(Forcings):
    """Retrieve boundary forcings from the input."""

    def __init__(self, context: Context, input: DatasetInput, variables: List[str], variables_mask: IntArray):
        """Initialize the BoundaryForcings object.

        Parameters
        ----------
        context : Context
            The context for the forcings.
        input : DatasetInput
            The input object.
        variables : List[str]
            The list of variables to retrieve.
        variables_mask : IntArray
            The mask to apply to the forcings.
        """
        super().__init__(context)
        self.variables = variables
        self.variables_mask = variables_mask
        assert isinstance(input, DatasetInput), "Currently only boundary forcings from dataset supported."
        self.input = input
        if "output_mask" in context.checkpoint._supporting_arrays:
            self.spatial_mask = ~context.checkpoint.load_supporting_array("output_mask")
        else:
            self.spatial_mask = np.array([False] * len(input["latitudes"]), dtype=bool)
        self.kinds = dict(retrieved=True)  # Used for debugging

    def __repr__(self) -> str:
        """Return a string representation of the BoundaryForcings object."""
        return f"{self.__class__.__name__}({self.variables})"

    def load_forcings_array(self, dates: List[Date], current_state: State) -> FloatArray:
        """Load the boundary forcings for the given dates.

        Parameters
        ----------
        dates : List[Date]
            The list of dates for which to load the forcings.
        current_state : State
            The current state of the model.

        Returns
        -------
        FloatArray
            The loaded forcings as a numpy array.
        """
        data = self._state_to_numpy(
            self.input.load_forcings_state(variables=self.variables, dates=dates, current_state=current_state),
            self.variables,
            dates,
        )
        data = data[..., self.spatial_mask]

        expected_shape = (len(self.variables), len(dates), current_state["latitudes"][self.spatial_mask].size)
        assert data.shape == expected_shape, (data.shape, expected_shape)

        return data



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import datetime
import os
import sys

read_the_docs_build = os.environ.get("READTHEDOCS", None) == "True"

sys.path.insert(0, os.path.join(os.path.abspath(".."), "src"))


source_suffix = ".rst"
master_doc = "index"
pygments_style = "sphinx"
html_theme_options = {"logo_only": True}
html_logo = "_static/logo.png"


# -- Project information -----------------------------------------------------

project = "Anemoi Inference"

author = "Anemoi contributors"

year = datetime.datetime.now().year
if year == 2024:
    years = "2024"
else:
    years = "2024-%s" % (year,)

copyright = "%s, Anemoi contributors" % (years,)


try:
    from anemoi.inference._version import __version__

    release = __version__
except ImportError:
    release = "0.0.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx_rtd_theme",
    "nbsphinx",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinxarg.ext",
    "sphinxcontrib.autodoc_pydantic",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "'**.ipynb_checkpoints'"]

intersphinx_mapping = {
    "python": ("https://python.readthedocs.io/en/latest", None),
    "anemoi-docs": (
        "https://anemoi.readthedocs.io/en/latest/",
        ("../../../anemoi-docs/docs/_build/html/objects.inv", None),
    ),
    "anemoi-utils": (
        "https://anemoi-utils.readthedocs.io/en/latest/",
        ("../../anemoi-utils/docs/_build/html/objects.inv", None),
    ),
    "anemoi-datasets": (
        "https://anemoi-datasets.readthedocs.io/en/latest/",
        ("../../anemoi-datasets/docs/_build/html/objects.inv", None),
    ),
    "anemoi-models": (
        "https://anemoi-models.readthedocs.io/en/latest/",
        ("../../anemoi-models/docs/_build/html/objects.inv", None),
    ),
    "anemoi-training": (
        "https://anemoi-training.readthedocs.io/en/latest/",
        ("../../anemoi-training/docs/_build/html/objects.inv", None),
    ),
    "anemoi-inference": (
        "https://anemoi-inference.readthedocs.io/en/latest/",
        ("../../anemoi-inference/docs/_build/html/objects.inv", None),
    ),
    "anemoi-graphs": (
        "https://anemoi-graphs.readthedocs.io/en/latest/",
        ("../../anemoi-graphs/docs/_build/html/objects.inv", None),
    ),
    "anemoi-registry": (
        "https://anemoi-registry.readthedocs.io/en/latest/",
        ("../../anemoi-registry/docs/_build/html/objects.inv", None),
    ),
    "anemoi-transform": (
        "https://anemoi-transform.readthedocs.io/en/latest/",
        ("../../anemoi-transform/docs/_build/html/objects.inv", None),
    ),
    "anemoi-plugins": (
        "https://anemoi-plugins.readthedocs.io/en/latest/",
        ("../../anemoi-plugins/docs/_build/html/objects.inv", None),
    ),
}


# https://www.notion.so/Deepnote-Launch-Buttons-63c642a5e875463495ed2341e83a4b2a


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Make the page width match the browser width
html_theme_options = {
    "body_max_width": "none",
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["style.css"]


todo_include_todos = not read_the_docs_build

autodoc_member_order = "bysource"  # Keep file order


# https://autodoc-pydantic.readthedocs.io/en/stable/users/configuration.html

autodoc_pydantic_model_show_json = True
autodoc_pydantic_model_show_field_summary = False
autodoc_pydantic_model_member_order = "bysource"

html_context = {
    "display_github": True,
    "github_user": "ecmwf",
    "github_repo": "anemoi-inference",
    "github_version": "main/docs/",
}



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

from anemoi.inference.types import State

from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("none")
class NoneOutput(Output):
    """None output class."""

    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        pass



# (C) Copyright 2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

from anemoi.inference.types import DataRequest
from anemoi.inference.types import State

if TYPE_CHECKING:
    from anemoi.inference.context import Context


class Processor(ABC):
    """Abstract base class for processors.

    Parameters
    ----------
    context : Context
        The context in which the processor operates.
    """

    def __init__(self, context: "Context") -> None:
        self.context = context
        self.checkpoint = context.checkpoint

    def __repr__(self) -> str:
        """Return a string representation of the processor.

        Returns
        -------
        str
            The class name of the processor.
        """
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def process(self, state: State) -> State:
        """Process the given state.

        Parameters
        ----------
        state : State
            The state to be processed.

        Returns
        -------
        State
            The processed state.
        """
        pass

    def patch_data_request(self, data_request: DataRequest) -> DataRequest:
        """Override if a processor needs to patch the data request (e.g. mars or cds).

        Parameters
        ----------
        data_request : DataRequest
            The data request to be patched.

        Returns
        -------
        DataRequest
            The patched data request.
        """
        return data_request



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime
import logging
from abc import ABC
from abc import abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING
from typing import Optional

from anemoi.inference.types import State

if TYPE_CHECKING:
    from anemoi.inference.context import Context

LOG = logging.getLogger(__name__)


class Output(ABC):
    """Abstract base class for output mechanisms."""

    def __init__(
        self,
        context: "Context",
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
    ):
        """Initialize the Output object.

        Parameters
        ----------
        context : Context
            The context in which the output operates.
        output_frequency : Optional[int], optional
            The frequency at which to output states, by default None.
        write_initial_state : Optional[bool], optional
            Whether to write the initial state, by default None.
        """
        self.context = context
        self.checkpoint = context.checkpoint
        self.reference_date = None

        self._write_step_zero = write_initial_state
        self._output_frequency = output_frequency

    def __repr__(self) -> str:
        """Return a string representation of the Output object.

        Returns
        -------
        str
            String representation of the Output object.
        """
        return f"{self.__class__.__name__}()"

    def write_initial_state(self, state: State) -> None:
        """Write the initial state.

        Parameters
        ----------
        state : State
            The initial state to write.
        """
        state.setdefault("step", datetime.timedelta(0))
        if self.write_step_zero:
            self.write_step(state)

    def write_state(self, state: State) -> None:
        """Write the state.

        Parameters
        ----------
        state : State
            The state to write.
        """
        self.open(state)

        step = state["step"]
        if self.output_frequency is not None:
            if (step % self.output_frequency).total_seconds() != 0:
                return

        return self.write_step(state)

    @classmethod
    def reduce(cls, state: State) -> State:
        """Create a new state which is a projection of the original state on the last step in the multi-steps dimension.

        Parameters
        ----------
        state : State
            The original state.

        Returns
        -------
        State
            The reduced state.
        """
        reduced_state = state.copy()
        reduced_state["fields"] = {}
        for field, values in state["fields"].items():
            reduced_state["fields"][field] = values[-1, :]
        return reduced_state

    def open(self, state: State) -> None:
        """Open the output for writing.

        Parameters
        ----------
        state : State
            The state to open.
        """
        # Override this method when initialisation is needed
        pass

    def close(self) -> None:
        """Close the output."""
        pass

    @abstractmethod
    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state to write.
        """
        pass

    @cached_property
    def write_step_zero(self) -> bool:
        """Determine whether to write the initial state."""
        if self._write_step_zero is not None:
            return self._write_step_zero

        return self.context.write_initial_state

    @cached_property
    def output_frequency(self) -> Optional[datetime.timedelta]:
        """Get the output frequency."""
        from anemoi.utils.dates import as_timedelta

        if self._output_frequency is not None:
            return as_timedelta(self._output_frequency)

        if self.context.output_frequency is not None:
            return as_timedelta(self.context.output_frequency)

        return None

    def print_summary(self, depth: int = 0) -> None:
        """Print a summary of the output configuration.

        Parameters
        ----------
        depth : int, optional
            The indentation depth for the summary, by default 0.
        """
        LOG.info(
            "%s%s: output_frequency=%s write_initial_state=%s",
            " " * depth,
            self,
            self.output_frequency,
            self.write_step_zero,
        )


class ForwardOutput(Output):
    """Subclass of Output that forwards calls to other outputs.

    Subclass from this class to implement the desired behaviour of `output_frequency`
    which should only apply to leaves.
    """

    def __init__(
        self, context: "Context", output_frequency: Optional[int] = None, write_initial_state: Optional[bool] = None
    ):
        """Initialize the ForwardOutput object.

        Parameters
        ----------
        context : Context
            The context in which the output operates.
        output_frequency : Optional[int], optional
            The frequency at which to output states, by default None.
        write_initial_state : Optional[bool], optional
            Whether to write the initial state, by default None.
        """
        super().__init__(context, output_frequency=None, write_initial_state=write_initial_state)
        if self.context.output_frequency is not None:
            LOG.warning("output_frequency is ignored for '%s'", self.__class__.__name__)

    @cached_property
    def output_frequency(self) -> Optional[datetime.timedelta]:
        """Get the output frequency."""
        return None



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import List
from typing import Optional

import numpy as np

from anemoi.inference.input import Input
from anemoi.inference.inputs import create_input
from anemoi.inference.types import Date
from anemoi.inference.types import State

from . import input_registry

LOG = logging.getLogger(__name__)


@input_registry.register("cutout")
class Cutout(Input):
    """Combines one or more LAMs into a global source using cutouts."""

    def __init__(self, context, **sources: dict[str, dict]):
        """Create a cutout input from a list of sources.

        Parameters
        ----------
        context : dict
            The context runner.
        sources : dict of sources
            A dictionary of sources to combine.
        """
        super().__init__(context)

        self.sources: dict[str, Input] = {}
        self.masks: dict[str, np.ndarray] = {}
        for src, cfg in sources.items():
            mask = cfg.pop("mask", f"{src}/cutout_mask")
            self.sources[src] = create_input(context, cfg)
            self.masks[src] = self.sources[src].checkpoint.load_supporting_array(mask)

    def __repr__(self):
        return f"Cutout({self.sources})"

    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """

        LOG.info(f"Concatenating states from {self.sources}")
        sources = list(self.sources.keys())

        state = self.sources[sources[0]].create_input_state(date=date)
        for source in sources[1:]:
            mask = self.masks[source]
            _state = self.sources[source].create_input_state(date=date)

            state["latitudes"] = np.concatenate([state["latitudes"], _state["latitudes"][..., mask]], axis=-1)
            state["longitudes"] = np.concatenate([state["longitudes"], _state["longitudes"][..., mask]], axis=-1)
            for field, values in state["fields"].items():
                state["fields"][field] = np.concatenate([values, _state["fields"][field][..., mask]], axis=-1)

        return state

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variables to load.
        dates : List[Date]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        State
            The loaded forcings state.
        """

        sources = list(self.sources.keys())
        fields = self.sources[sources[0]].load_forcings_state(
            variables=variables, dates=dates, current_state=current_state
        )["fields"]
        for source in sources[1:]:
            mask = self.masks[source]
            _fields = self.sources[source].load_forcings_state(
                variables=variables, dates=dates, current_state=current_state
            )["fields"]
            for field in fields:
                fields[field] = np.concatenate([fields[field], _fields[field][..., mask]], axis=-1)

        current_state["fields"] |= fields
        return current_state



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from typing import Any

from anemoi.utils.registry import Registry

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context

input_registry = Registry(__name__)


def create_input(context: Context, config: Configuration) -> Any:
    """Create an input instance from the given context and configuration.

    Parameters
    ----------
    context : Context
        The context in which the input is created.
    config : Configuration
        The configuration for the input.

    Returns
    -------
    Any
        The created input instance.
    """
    return input_registry.from_config(config, context)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from typing import Any
from typing import Dict

from anemoi.utils.registry import Registry

task_registry = Registry(__name__)


def create_task(name: str, config: Dict[str, Any], global_config: Dict[str, Any]) -> Any:
    """Create a task instance based on the given configuration.

    Parameters
    ----------
    name : str
        The name of the task.
    config : Dict[str, Any]
        The configuration for the task.
    global_config : Dict[str, Any]
        The global configuration.

    Returns
    -------
    Any
        The created task instance.
    """
    return task_registry.from_config(config, name, global_config=global_config)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import base64
import logging
import os
import zlib
from typing import Any
from typing import Dict
from typing import Optional

import earthkit.data as ekd

from . import IndexTemplateProvider
from . import template_provider_registry

LOG = logging.getLogger(__name__)

# 1 - Get a GRIB with mars: retrieve,param=tp,levtype=sfc,type=fc,step=6,target=data.grib,grid=0.25/0.25
# 2 - grib_set -s edition=2,packingType=grid_ccsds data.grib data.grib2
# 3 - grib_set -d 0 data.grib2 out.grib
# 4 - python -c 'import base64, sys, zlib;print(base64.b64encode(zlib.compress(open(sys.argv[1], "rb").read())))' out.grib

# tp in grib2, 0.25/0.25 grid


@template_provider_registry.register("builtin")
class BuiltinTemplates(IndexTemplateProvider):
    """Builtin templates provider."""

    def __init__(self, manager: Any, index_path: Optional[str] = None) -> None:
        """Initialize the BuiltinTemplates instance.

        Parameters
        ----------
        manager : Any
            The manager instance.
        index_path : Optional[str], optional
            The path to the index file, by default None.
        """
        if index_path is None:
            index_path = os.path.join(os.path.dirname(__file__), "builtin.yaml")

        super().__init__(manager, index_path)

    def load_template(self, grib: str, lookup: Dict[str, Any]) -> Optional[ekd.Field]:
        """Load the template for the given GRIB and lookup.

        Parameters
        ----------
        grib : str
            The GRIB string.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        Optional[ekd.Field]
            The loaded template field if found, otherwise None.
        """
        import earthkit.data as ekd

        template = zlib.decompress(base64.b64decode(grib))
        return ekd.from_source("memory", template)[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
from contextlib import contextmanager
from functools import cached_property
from typing import Any
from typing import Dict
from typing import Generator
from typing import Tuple

from anemoi.utils.dates import as_datetime
from anemoi.utils.dates import frequency_to_timedelta as to_timedelta

from .protocol import MetadataProtocol

LOG = logging.getLogger(__name__)


@contextmanager
def patch_function(target: Any, attribute: str, replacement: Any) -> Generator[None, None, None]:
    """Context manager to temporarily replace an attribute of a target object.

    Parameters
    ----------
    target : object
        The target object whose attribute will be replaced.
    attribute : str
        The name of the attribute to replace.
    replacement : any
        The replacement value for the attribute.

    Returns
    -------
    Generator[None, None, None]
        The context manager.
    """
    original = getattr(target, attribute)
    setattr(target, attribute, replacement)
    try:
        yield
    finally:
        setattr(target, attribute, original)


class PatchMixin(MetadataProtocol):

    # `self` is a `Metadata` object

    def patch_metadata(
        self, supporting_arrays: Dict[str, Any], root: str, force: bool = False
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Patch the metadata with supporting arrays and root.

        Parameters
        ----------
        supporting_arrays : dict
            The supporting arrays to patch.
        root : str
            The root path for the supporting arrays.
        force : bool, optional
            Whether to force the patching, by default False.

        Returns
        -------
        tuple
            The patched metadata and supporting arrays.
        """
        dataset = self._metadata["dataset"]

        if (
            "variable_metadata" not in dataset
            or "supporting_arrays_paths" not in dataset
            or "sources" not in dataset
            or not not self._supporting_arrays
            or force
        ):
            self._patch_variable_metadata()
            self._supporting_arrays = self._patch_supporting_arrays(supporting_arrays, root)

        return self._metadata, self._supporting_arrays

    def _patch_variable_metadata(self) -> None:
        """Patch the variable metadata."""

        try:
            return self._patch_variable_metadata_open_dataset_1()
        except Exception:
            LOG.exception("_patch_variable_metadata_open_dataset_1 failed")

        try:
            return self._patch_variable_metadata_open_dataset_2()
        except Exception:
            LOG.exception("_patch_variable_metadata_open_dataset_2 failed")

    @cached_property
    def _from_zarr(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Open the dataset and fetch metadata and supporting arrays."""
        # We assume that the datasets are reachable via the content of
        # ~/.config/anemoi/settings.toml.

        from anemoi.datasets import open_dataset

        dataset = self._metadata["dataset"]
        arguments = dataset["arguments"]

        def _(x: Any) -> Any:
            if isinstance(x, dict):
                return {k: _(v) for k, v in x.items()}
            if isinstance(x, list):
                return [_(v) for v in x]

            if isinstance(x, str):
                if x.endswith(".zarr"):
                    x = os.path.basename(x)
                    x = os.path.splitext(x)[0]

            return x

        args, kwargs = _(arguments["args"]), _(arguments["kwargs"])

        LOG.info(f"Opening dataset with args {args} and kwargs {kwargs}")
        ds = open_dataset(*args, **kwargs)
        return ds.metadata(), ds.supporting_arrays()

    def _patch_variable_metadata_open_dataset_1(self) -> None:
        """Try to open the dataset and re-fetch metadata.

        Returns
        -------
        None
        """
        # Try to open the dataset(s) and re-fetch metadata.
        # In the checkpoint we keep track of the arguments used to open the dataset.

        # First attempt, try to open the dataset

        dataset = self._metadata["dataset"]
        metadata, _ = self._from_zarr

        # Update the metadata
        for k, v in metadata.items():
            if k not in dataset:
                LOG.info("Updating metadata key `%s` with value `%s`", k, v)
                dataset[k] = v

    def _patch_variable_metadata_open_dataset_2(self) -> None:
        """Fetch the metadata from the catalogue."""

        import anemoi.datasets.data.stores
        import numpy as np
        import zarr
        from anemoi.registry import Dataset

        start_date = as_datetime(self._metadata["dataset"]["start_date"])
        end_date = as_datetime(self._metadata["dataset"]["end_date"])
        frequency = to_timedelta(self._metadata["dataset"]["frequency"])
        dates = []

        while start_date <= end_date:
            dates.append(start_date)
            start_date += frequency

        def _open_zarr(name: str) -> zarr.hierarchy.Group:
            nonlocal dates
            name = os.path.basename(name)
            name = os.path.splitext(name)[0]
            entry = Dataset(name)

            root = zarr.group()
            root.attrs.update(entry.record["metadata"])
            values = 10
            variables = 10

            data = np.zeros(shape=(2, variables, 1, values))
            root.create_dataset(
                "data",
                # data=data,
                dtype=data.dtype,
                shape=(2, variables, 1, values),
                # compressor=None,
            )
            dates = np.array(dates, dtype="datetime64")
            root.create_dataset(
                "dates",
                data=dates,
                compressor=None,
            )
            root.create_dataset(
                "latitudes",
                shape=(values,),
                # compressor=None,
            )
            root.create_dataset(
                "longitudes",
                shape=(values,),
                # compressor=None,
            )

            return root

        with patch_function(anemoi.datasets.data.stores, "open_zarr", _open_zarr):
            self._patch_variable_metadata_open_dataset_1()

    def _patch_supporting_arrays(self, supporting_arrays: Dict[str, Any], root: str) -> Dict[str, Any]:
        """Patch the supporting arrays.

        Parameters
        ----------
        supporting_arrays : dict
            The supporting arrays to patch.
        root : str
            The root path for the supporting arrays.

        Returns
        -------
        dict
            The patched supporting arrays.
        """
        metadata, supporting_arrays = self._from_zarr

        # assert False, metadata['sources']

        LOG.info("Supporting arrays: '%s'", metadata["supporting_arrays"])
        LOG.info("Supporting arrays: '%s'", supporting_arrays)

        supporting_arrays_paths = {
            key: dict(
                path=f"{root}/{key}.numpy",
                shape=value.shape,
                dtype=str(value.dtype),
            )
            for key, value in supporting_arrays.items()
        }

        for k, v in supporting_arrays_paths.items():
            LOG.info("Saving supporting array `%s` to %s (shape=%s, dtype=%s)", k, v["path"], v["shape"], v["dtype"])

        self._metadata["supporting_arrays_paths"] = supporting_arrays_paths

        return supporting_arrays



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

import earthkit.data as ekd
from anemoi.transform.filters import filter_registry

from anemoi.inference.decorators import main_argument

from ..processor import Processor
from . import pre_processor_registry

LOG = logging.getLogger(__name__)


@pre_processor_registry.register("forward_transform_filter")
@main_argument("filter")
class ForwardTransformFilter(Processor):
    """A processor that applies a forward transform filter to the given fields.

    This class uses a specified filter from the filter registry to process
    fields and patch data requests.

    Attributes
    ----------
    filter : object
        The filter instance used for processing fields and patching data requests.
    """

    def __init__(self, context: Any, filter: str, **kwargs: Any) -> None:
        """Initialize the ForwardTransformFilter.

        Parameters
        ----------
        context : object
            The context in which the filter is being used.
        filter : str
            The name of the filter to be used.
        **kwargs : dict
            Additional keyword arguments to pass to the filter.
        """
        super().__init__(context)
        self.filter = filter_registry.create(filter, **kwargs)

    def process(self, fields: ekd.FieldList) -> ekd.FieldList:
        """Process the given fields using the forward filter.

        Parameters
        ----------
        fields : ekd.FieldList
            The fields to be processed.

        Returns
        -------
        ekd.FieldList
            The processed fields.
        """
        return self.filter.forward(fields)

    def patch_data_request(self, data_request: Any) -> Any:
        """Patch the data request using the filter.

        Parameters
        ----------
        data_request : object
            The data request to be patched.

        Returns
        -------
        object
            The patched data request.
        """
        return self.filter.patch_data_request(data_request)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import unittest
from pathlib import Path
from typing import Dict
from unittest.mock import patch

import pytest

import anemoi.inference.checkpoint
from anemoi.inference.runner import Runner

from .fake_metadata import FakeMetadata


@pytest.fixture(scope="session")
def fake_huggingface_repo(tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Create a fake huggingface repo download.

    Parameters
    ----------
    tmp_path_factory : pytest.TempPathFactory
        Factory for temporary directories.

    Returns
    -------
    Path
        Path to the fake huggingface repo.
    """
    tmp_dir = tmp_path_factory.mktemp("repo")
    fn = tmp_dir / "model.ckpt"
    fn.write_text("TESTING", encoding="utf-8")
    return tmp_dir


@pytest.fixture(scope="session")
def fake_huggingface_ckpt(tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Create a fake huggingface ckpt download.

        Factory for temporary directories.

    Parameters
    ----------
    tmp_path_factory : pytest.TempPathFactory
        Factory for temporary directories

    Returns
    -------
    pathlib.Path
        Path to the fake huggingface checkpoint.
    """
    tmp_dir = tmp_path_factory.mktemp("repo")
    fn = tmp_dir / "model.ckpt"
    fn.write_text("TESTING", encoding="utf-8")
    return fn


@patch("huggingface_hub.snapshot_download")
@pytest.mark.parametrize("ckpt", ["organisation/test_repo"])
def test_huggingface_repo_download_str(
    huggingface_mock: unittest.mock.Mock,
    monkeypatch: pytest.MonkeyPatch,
    ckpt: str,
    fake_huggingface_repo: Path,
) -> None:
    """Test downloading a huggingface repo using a string identifier.

    Parameters
    ----------
    huggingface_mock : unittest.mock.Mock
        Mock for the huggingface snapshot download.
    monkeypatch : pytest.MonkeyPatch
        Monkeypatch fixture for modifying attributes.
    ckpt : str
        Checkpoint identifier.
    fake_huggingface_repo : pathlib.Path
        Path to the fake huggingface repo.
    """
    monkeypatch.setattr(anemoi.inference.checkpoint.Checkpoint, "_metadata", FakeMetadata())
    huggingface_mock.return_value = fake_huggingface_repo

    runner = Runner({"huggingface": ckpt})
    assert runner.checkpoint.path == str(fake_huggingface_repo / "model.ckpt")

    assert huggingface_mock.called
    huggingface_mock.assert_called_once_with(repo_id=ckpt)


@patch("huggingface_hub.snapshot_download")
@pytest.mark.parametrize("ckpt", [{"repo_id": "organisation/test_repo"}])
def test_huggingface_repo_download_dict(
    huggingface_mock: unittest.mock.Mock,
    monkeypatch: pytest.MonkeyPatch,
    ckpt: Dict[str, str],
    fake_huggingface_repo: Path,
) -> None:
    """Test downloading a huggingface repo using a dictionary identifier.

    Parameters
    ----------
    huggingface_mock : unittest.mock.Mock
        Mock for the huggingface snapshot download.
    monkeypatch : pytest.MonkeyPatch
        Monkeypatch fixture for modifying attributes.
    ckpt : Dict[str, str]
        Checkpoint identifier.
    fake_huggingface_repo : pathlib.Path
        Path to the fake huggingface repo.
    """
    monkeypatch.setattr(anemoi.inference.checkpoint.Checkpoint, "_metadata", FakeMetadata())
    huggingface_mock.return_value = fake_huggingface_repo

    runner = Runner({"huggingface": ckpt})
    assert runner.checkpoint.path == str(fake_huggingface_repo / "model.ckpt")

    assert huggingface_mock.called
    huggingface_mock.assert_called_once_with(**ckpt)


@patch("huggingface_hub.hf_hub_download")
@pytest.mark.parametrize("ckpt", [{"repo_id": "organisation/test_repo", "filename": "model.ckpt"}])
def test_huggingface_file_download(
    huggingface_mock: unittest.mock.Mock,
    monkeypatch: pytest.MonkeyPatch,
    ckpt: Dict[str, str],
    fake_huggingface_ckpt: Path,
) -> None:
    """Test downloading a specific file from a huggingface repo.

    Parameters
    ----------
    huggingface_mock : unittest.mock.Mock
        Mock for the huggingface file download.
    monkeypatch : pytest.MonkeyPatch
        Monkeypatch fixture for modifying attributes.
    ckpt : Dict[str, str]
        Checkpoint identifier.
    fake_huggingface_ckpt : pathlib.Path
        Path to the fake huggingface checkpoint.
    """
    monkeypatch.setattr(anemoi.inference.checkpoint.Checkpoint, "_metadata", FakeMetadata())
    huggingface_mock.return_value = fake_huggingface_ckpt

    runner = Runner({"huggingface": ckpt})
    assert runner.checkpoint.path == str(fake_huggingface_ckpt)

    assert huggingface_mock.called
    huggingface_mock.assert_called_once_with(**ckpt)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
from argparse import ArgumentParser
from argparse import Namespace

from ..config.run import RunConfiguration
from ..runners import create_runner
from . import Command

LOG = logging.getLogger(__name__)


class RunCmd(Command):
    """Run inference from a config yaml file."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("--defaults", action="append", help="Sources of default values.")
        command_parser.add_argument(
            "config",
            help="Path to config file. Can be omitted to pass config with overrides and defaults.",
        )
        command_parser.add_argument("overrides", nargs="*", help="Overrides as key=value")

    def run(self, args: Namespace) -> None:
        """Run the inference command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        if "=" in args.config:
            args.overrides.append(args.config)
            args.config = {}

        config = RunConfiguration.load(
            args.config,
            args.overrides,
            defaults=args.defaults,
        )

        runner = create_runner(config)
        runner.execute()


command = RunCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""List of precisions supported by the inference runner."""


import torch

PRECISIONS: dict[str, torch.dtype] = {
    "16-mixed": torch.float16,
    "16": torch.float16,
    "32": torch.float32,
    "b16-mixed": torch.bfloat16,
    "b16": torch.bfloat16,
    "bf16-mixed": torch.bfloat16,
    "bf16": torch.bfloat16,
    "bfloat16": torch.bfloat16,
    "f16": torch.float16,
    "f32": torch.float32,
    "float16": torch.float16,
    "float32": torch.float32,
}



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""This is an `ai-models` plugin that can run anemoi.inference models."""

import argparse
import logging
import os
from functools import cached_property
from typing import Any
from typing import List
from typing import Optional

from ai_models.model import Model

from anemoi.inference.inputs.grib import GribInput
from anemoi.inference.outputs.grib import GribOutput
from anemoi.inference.runner import PRECISIONS as AUTOCAST
from anemoi.inference.runners.plugin import PluginRunner
from anemoi.inference.types import Date
from anemoi.inference.types import State

LOG = logging.getLogger(__name__)


class FieldListInput(GribInput):
    """Handles earchkit-data fieldlists input fields."""

    def __init__(self, context: Any, *, input_fields: Any) -> None:
        """Initialize FieldListInput.

        Parameters
        ----------
        context : Any
            The context for the input.
        input_fields : Any
            The input fields to be processed.
        """
        super().__init__(context)
        self.input_fields = input_fields

    def create_input_state(self, *, date: Optional[Date]) -> Any:
        """Create the input state for the given date.

        Parameters
        ----------
        date : str
            The date for which to create the input state.

        Returns
        -------
        Any
            The created input state.
        """
        return self._create_input_state(self.input_fields, variables=None, date=date)

    def load_forcings_state(
        self,
        *,
        variables: List[str],
        dates: List[str],
        current_state: State,
    ) -> State:
        """Load the forcings state.

        Parameters
        ----------
        variables : List[str]
            The variables to load.
        dates : List[str]
            The dates for which to load the forcings.
        current_state : State
            The current state to update.

        Returns
        -------
        State
            The updated state with loaded forcings.
        """
        return self._load_forcings_state(
            self.input_fields,
            variables=variables,
            dates=dates,
            current_state=current_state,
        )

    def set_private_attributes(self, state: State, input_fields: Any) -> None:
        """Set private attributes for the state.

        Parameters
        ----------
        state : Dict[str, Any]
            The state to update.
        input_fields : Any
            The input fields to use for setting attributes.
        """
        input_fields = input_fields.order_by("valid_datetime")
        state["_grib_templates_for_output"] = {field.metadata("name"): field for field in input_fields}


class CallbackOutput(GribOutput):
    """Call ai-models write method."""

    def __init__(self, context: Any, *, write: Any, encoding: Any = None) -> None:
        """Initialize CallbackOutput.

        Parameters
        ----------
        context : Any
            The context for the output.
        write : Any
            The write method to call.
        encoding : Any, optional
            The encoding to use, by default None.
        """
        super().__init__(context, encoding=encoding, templates={"source": "templates"})
        self.write = write

    def write_message(self, message: Any, *args: Any, **kwargs: Any) -> None:
        """Write a message using the write method.

        Parameters
        ----------
        message : Any
            The message to write.
        args : Any
            Additional arguments.
        kwargs : Any
            Additional keyword arguments.
        """
        self.write(message, *args, **kwargs)


class AIModelPlugin(Model):

    expver: Any = None

    def add_model_args(self, parser: argparse.ArgumentParser) -> None:
        """To be implemented in subclasses to add model-specific arguments to the parser.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            An instance of the parser to add arguments to.
        """
        pass

    def parse_model_args(self, args: List[str]) -> argparse.ArgumentParser:
        """Parse model-specific arguments.

        Parameters
        ----------
        args : List[str]
            The list of arguments to parse.

        Returns
        -------
        argparse.ArgumentParser
            The parser with parsed arguments.
        """
        parser = argparse.ArgumentParser()

        parser.add_argument("--checkpoint", required=not hasattr(self, "download_files"))
        parser.add_argument(
            "--autocast",
            type=str,
            choices=sorted(AUTOCAST.keys()),
        )

        self.add_model_args(parser)

        args = parser.parse_args(args)
        args._checkpoint = args.checkpoint

        if args._checkpoint is None:
            args._checkpoint = os.path.join(self.assets, self.download_files[0])

        for k, v in vars(args).items():
            setattr(self, k, v)

        return parser

    @cached_property
    def runner(self) -> PluginRunner:
        """Get the PluginRunner instance."""
        return PluginRunner(self._checkpoint, device=self.device)

    def run(self) -> None:
        """Run the model inference."""
        if self.deterministic:
            self.torch_deterministic_mode()

        input_kwargs = self.input.anemoi_plugin_input_kwargs()
        output_kwargs = self.input.anemoi_plugin_input_kwargs()

        input = FieldListInput(self.runner, input_fields=self.all_fields, **input_kwargs)
        output = CallbackOutput(self.runner, write=self.write, **output_kwargs)

        input_state = input.create_input_state(date=self.start_datetime)

        output.write_initial_state(input_state)

        for state in self.runner.run(input_state=input_state, lead_time=self.lead_time):
            output.write_state(state)

        output.close()

    # Below are methods forwarded to the checkpoint

    @property
    def param_sfc(self) -> Any:
        """Surface parameters."""
        return self.runner.param_sfc

    @property
    def param_level_pl(self) -> Any:
        """Pressure level parameters."""
        return self.runner.param_level_pl

    @property
    def param_level_ml(self) -> Any:
        """Model level parameters."""
        return self.runner.param_level_ml

    @property
    def constant_fields(self) -> Any:
        """Constant fields from input."""
        return self.runner.checkpoint.constants_from_input

    @property
    def grid(self) -> Any:
        """Grid information."""
        return self.runner.checkpoint.grid

    @property
    def area(self) -> Any:
        """Area information."""
        return self.runner.checkpoint.area

    @property
    def lagged(self) -> Any:
        """Lagged information."""
        return self.runner.lagged


model = AIModelPlugin



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import FloatArray
from anemoi.inference.types import State

from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)

ListOrAll = Union[List[str], Literal["all"]]


def fix(lons: FloatArray) -> FloatArray:
    """Fix longitudes greater than 180 degrees.

    Parameters
    ----------
    lons : FloatArray
        Array of longitudes.

    Returns
    -------
    FloatArray
        Fixed array of longitudes.
    """
    return np.where(lons > 180, lons - 360, lons)


@output_registry.register("plot")
class PlotOutput(Output):
    """Plot output class."""

    def __init__(
        self,
        context: Context,
        path: str,
        variables: ListOrAll = "all",
        strftime: str = "%Y%m%d%H%M%S",
        template: str = "plot_{variable}_{date}.{format}",
        dpi: int = 300,
        format: str = "png",
        missing_value: Optional[float] = None,
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
    ) -> None:
        """Initialize the PlotOutput.

        Parameters
        ----------
        context : Context
            The context.
        path : str
            The path to save the plots.
        variables : list, optional
            The list of variables to plot, by default all.
        strftime : str, optional
            The date format string, by default "%Y%m%d%H%M%S".
        template : str, optional
            The template for plot filenames, by default "plot_{variable}_{date}.{format}".
        dpi : int, optional
            The resolution of the plot, by default 300.
        format : str, optional
            The format of the plot, by default "png".
        missing_value : float, optional
            The value to use for missing data, by default None.
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        """

        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self.path = path
        self.format = format
        self.variables = variables
        self.strftime = strftime
        self.template = template
        self.dpi = dpi
        self.missing_value = missing_value

        if self.variables != "all":
            if not isinstance(self.variables, (list, tuple)):
                self.variables = [self.variables]

    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        import cartopy.crs as ccrs
        import cartopy.feature as cfeature
        import matplotlib.pyplot as plt
        import matplotlib.tri as tri

        os.makedirs(self.path, exist_ok=True)

        longitudes = state["longitudes"]
        latitudes = state["latitudes"]
        triangulation = tri.Triangulation(fix(longitudes), latitudes)

        for name, values in state["fields"].items():

            if self.variables != "all" and name not in self.variables:
                continue

            _, ax = plt.subplots(subplot_kw={"projection": ccrs.PlateCarree()})
            ax.coastlines()
            ax.add_feature(cfeature.BORDERS, linestyle=":")

            missing_values = np.isnan(values)
            missing_value = self.missing_value
            if missing_value is None:
                min = np.nanmin(values)
                missing_value = min - np.abs(min) * 0.001

            values = np.where(missing_values, self.missing_value, values)

            _ = ax.tricontourf(triangulation, values, levels=10, transform=ccrs.PlateCarree())

            ax.tricontour(
                triangulation,
                values,
                levels=10,
                colors="black",
                linewidths=0.5,
                transform=ccrs.PlateCarree(),
            )

            date = state["date"].strftime("%Y-%m-%d %H:%M:%S")
            ax.set_title(f"{name} at {date}")

            date = state["date"].strftime(self.strftime)
            fname = self.template.format(date=date, variable=name, format=self.format)
            fname = os.path.join(self.path, fname)

            plt.savefig(fname, dpi=self.dpi, bbox_inches="tight")
            plt.close()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging


def test_inputs() -> None:
    """Test the inputs module."""
    pass


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import re
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import earthkit.data as ekd
import numpy as np
from earthkit.data.indexing.fieldlist import FieldArray
from earthkit.data.utils.dates import to_datetime
from numpy.typing import DTypeLike

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import FloatArray
from anemoi.inference.types import State

from ..checks import check_data
from ..input import Input

LOG = logging.getLogger(__name__)


class RulesNamer:
    """A namer that uses rules to generate names."""

    def __init__(self, rules: Any, default_namer: Callable[[Any, Dict[str, Any]], str]) -> None:
        """Initialize the RulesNamer.

        Parameters
        ----------
        rules : List[List[Dict[str, Any], Dict[str, Any]]]
            The rules for naming.
        default_namer : Callable[[Any, Dict[str, Any]], str]
            The default namer to use if no rules match.
        """
        self.rules = rules
        self.default_namer = default_namer

    def __call__(self, field: Any, original_metadata: Dict[str, Any]) -> str:
        """Generate a name for the field.

        Parameters
        ----------
        field : Any
            The field for which to generate a name.
        original_metadata : Dict[str, Any]
            The original metadata of the field.

        Returns
        -------
        str
            The generated name.
        """
        for rule in self.rules:
            assert len(rule) == 2, rule
            ok = True
            for k, v in rule[0].items():
                if original_metadata.get(k) != v:
                    ok = False
            if ok:
                return self.substitute(rule[1], field, original_metadata)

        return self.default_namer(field, original_metadata)

    def substitute(self, template: str, field: Any, original_metadata: Dict[str, Any]) -> str:
        """Substitute placeholders in the template with metadata values.

        Parameters
        ----------
        template : str
            The template string with placeholders.
        field : Any
            The field for which to generate a name.
        original_metadata : Dict[str, Any]
            The original metadata of the field.

        Returns
        -------
        str
            The generated name with placeholders substituted.
        """
        matches = re.findall(r"\{(.+?)\}", template)
        matches = {m: original_metadata.get(m) for m in matches}
        return template.format(**matches)


class EkdInput(Input):
    """Handles earthkit-data FieldList as input."""

    def __init__(
        self,
        context: Context,
        *,
        namer: Optional[Union[Callable[[Any, Dict[str, Any]], str], Dict[str, Any]]] = None,
    ) -> None:
        """Initialize the EkdInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        namer : Optional[Union[Callable[[Any, Dict[str, Any]], str], Dict[str, Any]]]
            Optional namer for the input.
        """
        super().__init__(context)

        if isinstance(namer, dict):
            # TODO: a factory for namers
            assert "rules" in namer, namer
            assert len(namer) == 1, namer
            namer = RulesNamer(namer["rules"], self.checkpoint.default_namer())

        self._namer = namer if namer is not None else self.checkpoint.default_namer()
        assert callable(self._namer), type(self._namer)

    def _filter_and_sort(self, data: Any, *, variables: List[str], dates: List[Any], title: str) -> Any:
        """Filter and sort the data.

        Parameters
        ----------
        data : Any
            The data to filter and sort.
        variables : List[str]
            The list of variables to select.
        dates : List[Any]
            The list of dates to select.
        title : str
            The title for logging.

        Returns
        -------
        Any
            The filtered and sorted data.
        """

        def _name(field: Any, _: Any, original_metadata: Dict[str, Any]) -> str:
            return self._namer(field, original_metadata)

        data = FieldArray([f.clone(name=_name) for f in data])

        valid_datetime = [_.isoformat() for _ in dates]
        LOG.info("Selecting fields %s %s", len(data), valid_datetime)

        # for f in data:
        #     LOG.info("Field %s %s", f.metadata("name"), f.metadata("valid_datetime"))

        data = data.sel(name=variables, valid_datetime=valid_datetime).order_by(
            name=variables, valid_datetime="ascending"
        )

        check_data(title, data, variables, dates)

        return data

    def _find_variable(self, data: Any, name: str, **kwargs: Any) -> Any:
        """Find a variable in the data.

        Parameters
        ----------
        data : Any
            The data to search.
        name : str
            The name of the variable to find.
        **kwargs : Any
            Additional arguments for selecting the variable.

        Returns
        -------
        Any
            The selected variable.
        """

        def _name(field: Any, _: Any, original_metadata: Dict[str, Any]) -> str:
            return self._namer(field, original_metadata)

        data = FieldArray([f.clone(name=_name) for f in data])
        return data.sel(name=name, **kwargs)

    def _create_state(
        self,
        fields: ekd.FieldList,
        *,
        variables: Optional[List[str]] = None,
        dates: List[Date],
        latitudes: Optional[FloatArray] = None,
        longitudes: Optional[FloatArray] = None,
        dtype: DTypeLike = np.float32,
        flatten: bool = True,
        title: str = "Create state",
    ) -> State:
        """Create a state from an ekd.FieldList.

        Parameters
        ----------
        fields : ekd.FieldList
            The ekd fields.
        variables : Optional[List[str]]
            List of variables.
        dates : List[Date]
            The dates for which to create the input state.
        latitudes : Optional[FloatArray]
            The latitudes.
        longitudes : Optional[FloatArray]
            The longitudes.
        dtype : DTypeLike
            The data type.
        flatten : bool
            Whether to flatten the data.
        title : str
            The title for logging.

        Returns
        -------
        State
            The created input state.
        """
        for processor in self.context.pre_processors:
            LOG.info("Processing with %s", processor)
            fields = processor.process(fields)

        if variables is None:
            variables = self.checkpoint.variables_from_input(include_forcings=True)

        if len(fields) == 0:
            raise ValueError("No input fields provided")

        dates = sorted([to_datetime(d) for d in dates])
        date_to_index = {d.isoformat(): i for i, d in enumerate(dates)}

        state = dict(date=dates[-1], latitudes=latitudes, longitudes=longitudes, fields=dict())

        state_fields = state["fields"]

        fields = self._filter_and_sort(fields, variables=variables, dates=dates, title="Create input state")

        if latitudes is None and longitudes is None:
            try:
                state["latitudes"], state["longitudes"] = fields[0].grid_points()
                LOG.info(
                    "%s: using `latitudes` and `longitudes` from the first input field",
                    self.__class__.__name__,
                )
            except Exception as e:
                LOG.info(
                    "%s: could not get `latitudes` and `longitudes` from the input fields.",
                    self.__class__.__name__,
                )
                latitudes = self.checkpoint.latitudes
                longitudes = self.checkpoint.longitudes
                if latitudes is not None and longitudes is not None:
                    state["latitudes"] = latitudes
                    state["longitudes"] = longitudes
                    LOG.info(
                        "%s: using `latitudes` and `longitudes` found in the checkpoint.",
                        self.__class__.__name__,
                    )
                else:
                    LOG.error(
                        "%s: could not find `latitudes` and `longitudes` in the input fields or the checkpoint.",
                        self.__class__.__name__,
                    )
                    raise e

        check = defaultdict(set)

        n_points = fields[0].to_numpy(dtype=dtype, flatten=flatten).size
        for field in fields:

            name, valid_datetime = field.metadata("name"), field.metadata("valid_datetime")
            if name not in state_fields:
                state_fields[name] = np.full(
                    shape=(len(dates), n_points),
                    fill_value=np.nan,
                    dtype=dtype,
                )

            date_idx = date_to_index[valid_datetime]

            try:
                state_fields[name][date_idx] = field.to_numpy(dtype=dtype, flatten=flatten)
            except ValueError:
                LOG.error(
                    "Error with field %s: expected shape=%s, got shape=%s", name, state_fields[name].shape, field.shape
                )
                LOG.error("dates %s", dates)
                LOG.error("number_of_grid_points %s", self.checkpoint.number_of_grid_points)
                raise

            if date_idx in check[name]:
                LOG.error("Duplicate dates for %s: %s", name, date_idx)
                LOG.error("Expected %s", list(date_to_index.keys()))
                LOG.error("Got %s", list(check[name]))
                raise ValueError(f"Duplicate dates for {name}")

            check[name].add(date_idx)

        for name, idx in check.items():
            if len(idx) != len(dates):
                LOG.error("Missing dates for %s: %s", name, idx)
                LOG.error("Expected %s", list(date_to_index.keys()))
                LOG.error("Got %s", list(idx))
                raise ValueError(f"Missing dates for {name}")

        if self.context.trace:
            for name in check.keys():
                self.context.trace.from_input(name, self)

        # This is our chance to communicate output object
        # This is useful for GRIB that requires a template field
        # to be used as output
        self.set_private_attributes(state, fields)

        return state

    def _create_input_state(
        self,
        input_fields: ekd.FieldList,
        *,
        date: Optional[Date] = None,
        variables: Optional[List[str]] = None,
        latitudes: Optional[FloatArray] = None,
        longitudes: Optional[FloatArray] = None,
        dtype: DTypeLike = np.float32,
        flatten: bool = True,
    ) -> State:
        """Create the input state.

        Parameters
        ----------
        input_fields : ekd.FieldList
            The input fields.
        date : Date
            The date for which to create the input state.
        variables : Optional[List[str]]
            List of variables.
        latitudes : Optional[FloatArray]
            The latitudes.
        longitudes : Optional[FloatArray]
            The longitudes.
        dtype : DTypeLike
            The data type.
        flatten : bool
            Whether to flatten the data.

        Returns
        -------
        State
            The created input state.
        """
        if date is None:
            date = input_fields.order_by(valid_datetime="ascending")[-1].datetime()["valid_time"]
            LOG.info(
                "%s: `date` not provided, using the most recent date: %s", self.__class__.__name__, date.isoformat()
            )

        # TODO: where we do this might change in the future
        date = to_datetime(date)
        dates = [date + h for h in self.checkpoint.lagged]

        return self._create_state(
            input_fields,
            variables=variables,
            dates=dates,
            latitudes=latitudes,
            longitudes=longitudes,
            dtype=dtype,
            flatten=flatten,
            title="Create input state",
        )

    def _load_forcings_state(
        self,
        fields: ekd.FieldList,
        *,
        variables: List[str],
        dates: List[Date],
        current_state: State,
    ) -> State:
        """Load the forcings state.

        Parameters
        ----------
        fields : ekd.FieldList
            The fields to load.
        variables : List[str]
            The list of variables to load.
        dates : List[Any]
            The list of dates to load.
        current_state : Dict[str, Any]
            The current state.

        Returns
        -------
        State
            The loaded forcings state.
        """
        for processor in self.context.pre_processors:
            LOG.info("Processing with %s", processor)
            fields = processor.process(fields)

        return self._create_state(
            fields,
            variables=variables,
            dates=dates,
            latitudes=current_state["latitudes"],
            longitudes=current_state["longitudes"],
            dtype=np.float32,
            flatten=True,
            title="Load forcings state",
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import queue
import threading
from typing import Any
from typing import Dict
from typing import Optional

from anemoi.utils.logs import set_logging_name

from anemoi.inference.task import Task
from anemoi.inference.types import State

from ..transport import Transport
from . import transport_registry

LOG = logging.getLogger(__name__)


class TaskWrapper:
    """Wraps a task to be executed in a thread."""

    def __init__(self, task: Task) -> None:
        """Initialize the TaskWrapper.

        Parameters
        ----------
        task : Task
            The task to be wrapped.
        """
        self.task: Task = task
        self.queue: queue.Queue[Any] = queue.Queue(maxsize=1)
        self.error: Optional[Exception] = None
        self.name: str = task.name

    def run(self, transport: "ThreadsTransport") -> None:
        """Run the task within the given transport.

        Parameters
        ----------
        transport : ThreadsTransport
            The transport in which the task is run.
        """
        set_logging_name(self.task.name)
        try:
            self.task.run(transport)
        except Exception as e:
            LOG.exception(e)
            self.error = e

    def __repr__(self) -> str:
        """Return a string representation of the TaskWrapper.

        Returns
        -------
        str
            String representation of the TaskWrapper.
        """
        return repr(self.task)


@transport_registry.register("threads")
class ThreadsTransport(Transport):
    """Transport implementation using threads."""

    def __init__(self, couplings: Any, tasks: Dict[str, Task], *args: Any, **kwargs: Any) -> None:
        """Initialize the ThreadsTransport.

        Parameters
        ----------
        couplings : Any
            The couplings for the transport.
        tasks : Dict[str, Any]
            The tasks to be executed.
        """
        super().__init__(couplings, tasks)
        self.threads: Dict[str, threading.Thread] = {}
        self.lock = threading.Lock()
        self.backlogs: Dict[str, Any] = {name: {} for name in tasks}

    def start(self) -> None:
        """Start the transport by initializing and starting threads for each task."""
        self.wrapped_tasks = {name: TaskWrapper(task) for name, task in self.tasks.items()}

        for name, wrapped_task in self.wrapped_tasks.items():
            self.threads[name] = threading.Thread(target=wrapped_task.run, args=(self,))
            self.threads[name].start()

    def wait(self) -> None:
        """Wait for all threads to complete and handle any errors."""
        # TODO: wait for all threads, and kill remaining threads if any of them failed
        for name, thread in self.threads.items():
            thread.join()
            LOG.info(f"Thread `{name}` finished")

        for name, wrapped_task in self.wrapped_tasks.items():
            if wrapped_task.error:
                raise wrapped_task.error

    def send(self, sender: Task, target: Task, state: State, tag: int) -> None:
        """Send a state from the sender to the target.

        Parameters
        ----------
        sender : Any
            The task sending the state.
        target : Any
            The task receiving the state.
        state : Any
            The state to be sent.
        tag : int
            The tag associated with the state.
        """
        self.wrapped_tasks[target.name].queue.put((sender.name, tag, state.copy()))

    def receive(self, receiver: Task, source: Task, tag: int) -> State:
        """Receive a state from the source to the receiver.

        Parameters
        ----------
        receiver : Any
            The task receiving the state.
        source : Any
            The task sending the state.
        tag : int
            The tag associated with the state.

        Returns
        -------
        Any
            The received state.
        """
        LOG.info(f"{receiver}: receiving from {source} [{tag}] (backlog: {len(self.backlogs[receiver.name])})")

        if (source.name, tag) in self.backlogs[receiver.name]:
            with self.lock:
                return self.backlogs[receiver.name].pop((source.name, tag))

        while True:
            (sender, tag, state) = self.wrapped_tasks[receiver.name].queue.get()
            if sender != source.name or tag != tag:
                with self.lock:
                    self.backlogs[receiver.name][(sender, tag)] = state
                continue

            return state



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import earthkit.data as ekd

from anemoi.inference.types import State

from . import create_template_provider

LOG = logging.getLogger(__name__)


class TemplateManager:
    """A class to manage GRIB templates."""

    def __init__(self, owner: Any, templates: Optional[Union[List[str], str]] = None) -> None:
        """Initialize the TemplateManager.

        Parameters
        ----------
        owner : Any
            The owner of the TemplateManager.
        templates : Optional[Union[List[str], str]], optional
            A list of template names or a single template name, by default None.
        """
        self.owner = owner
        self.checkpoint = owner.context.checkpoint
        self.typed_variables = self.checkpoint.typed_variables

        self._template_cache = {}

        if templates is None:
            templates = []

        if not isinstance(templates, (list, tuple)):
            templates = [templates]

        if len(templates) == 0:
            templates = ["builtin"]

        self.templates_providers = [create_template_provider(self, template) for template in templates]

    def template(self, name: str, state: State) -> Optional[ekd.Field]:
        """Get the template for a given name and state.

        Parameters
        ----------
        name : str
            The name of the template.
        state : Dict[str, Any]
            The state dictionary containing template information.

        Returns
        -------
        Optional[ekd.Field]
            The template field if found, otherwise None.
        """
        assert name is not None, name

        # Use input fields as templates
        self._template_cache.update(state.get("_grib_templates_for_output", {}))

        if name not in self._template_cache:
            self.load_template(name, state)

        return self._template_cache.get(name)

    def load_template(self, name: str, state: State) -> Optional[ekd.Field]:
        """Load the template for a given name and state.

        Parameters
        ----------
        name : str
            The name of the template.
        state : Dict[str, Any]
            The state dictionary containing template information.

        Returns
        -------
        Optional[ekd.Field]
            The template field if found, otherwise None.
        """

        checkpoint = self.owner.context.checkpoint

        typed = checkpoint.typed_variables[name]

        lookup = dict(
            name=name,
            grid=self._grid(checkpoint.grid),
            time_processing=typed.time_processing,
            number_of_grid_points=checkpoint.number_of_grid_points,
        )

        for key, value in typed.grib_keys.items():
            if key in ("step", "date", "time", "hdate"):
                continue
            lookup[key] = value

        lookup.update(self.owner.template_lookup(name))

        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug(f"Loading template for `{name}` with lookup:")
            LOG.debug("%s", json.dumps(lookup, indent=2, default=str))

        tried = []
        for provider in self.templates_providers:
            template = provider.template(name, lookup)
            if template is not None:
                self._template_cache[name] = template
                return

            tried.append(provider)

        LOG.warning(f"Could not find template for `{name}` in {tried}")
        LOG.warning(f"Loading template for `{name}` with lookup:")
        LOG.warning("%s", json.dumps(lookup, indent=2, default=str))
        return None

    def _grid(self, grid: Union[str, List[float], Tuple[int, int]]) -> Union[str, List[float]]:
        """Convert the grid information to a standardized format.

        Parameters
        ----------
        grid : Union[str, List[int], Tuple[int, int]]
            The grid information.

        Returns
        -------
        Union[str, int]
            The standardized grid format.
        """

        if isinstance(grid, str):
            return grid.upper()

        if isinstance(grid, (tuple, list)) and len(grid) == 2:
            if grid[0] == grid[1]:
                return grid[0]
            return f"{grid[0]}x{grid[1]}"

        return grid



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Optional

import numpy as np

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context
from anemoi.inference.types import State

from ..output import ForwardOutput
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("extract_lam")
class ExtractLamOutput(ForwardOutput):
    """Extract LAM output class.

    Parameters
    ----------
    context : dict
        The context dictionary.
    output : dict
        The output configuration dictionary.
    lam : str, optional
        The LAM identifier, by default "lam_0".
    output_frequency : int, optional
        The frequency of output, by default None.
    write_initial_state : bool, optional
        Whether to write the initial state, by default None.
    """

    def __init__(
        self,
        context: Context,
        *,
        output: Configuration,
        lam: str = "lam_0",
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
    ) -> None:
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)

        if "cutout_mask" in self.checkpoint.supporting_arrays:
            # Backwards compatibility
            mask = self.checkpoint.load_supporting_array("cutout_mask")
            points = slice(None, -np.sum(mask))
        else:
            if "lam_0" not in lam:
                raise NotImplementedError("Only lam_0 is supported")

            if "lam_1/cutout_mask" in self.checkpoint.supporting_arrays:
                raise NotImplementedError("Only lam_0 is supported")

            mask = self.checkpoint.load_supporting_array(f"{lam}/cutout_mask")
            assert len(mask) == np.sum(mask)
            points = slice(None, np.sum(mask))

        self.points = points
        self.output = create_output(context, output)

    def __repr__(self) -> str:
        """Return a string representation of the ExtractLamOutput object."""
        return f"ExtractLamOutput({self.points}, {self.output})"

    def write_initial_state(self, state: State) -> None:
        """Write the initial step of the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_initial_state(self._apply_mask(state))

    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_state(self._apply_mask(state))

    def _apply_mask(self, state: State) -> State:
        """Apply the mask to the state.

        Parameters
        ----------
        state : State
            The state dictionary.

        Returns
        -------
        State
            The masked state dictionary.
        """
        state = state.copy()
        state["fields"] = state["fields"].copy()
        state["latitudes"] = state["latitudes"][self.points]
        state["longitudes"] = state["longitudes"][self.points]

        for field in state["fields"]:
            data = state["fields"][field]
            if data.ndim == 1:
                data = data[self.points]
            else:
                data = data[..., self.points]
            state["fields"][field] = data

        return state

    def close(self) -> None:
        """Close the output."""
        self.output.close()

    def print_summary(self, depth: int = 0) -> None:
        """Print the summary of the output.

        Parameters
        ----------
        depth : int, optional
            The depth of the summary, by default 0.
        """
        super().print_summary(depth)
        self.output.print_summary(depth + 1)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Any
from typing import Dict

import earthkit.data as ekd

from . import TemplateProvider
from . import template_provider_registry

LOG = logging.getLogger(__name__)


@template_provider_registry.register("file")
class FileTemplates(TemplateProvider):
    """Template provider using a single GRIB file."""

    def __init__(self, manager: Any, path: str) -> None:
        """Initialize the FileTemplates instance.

        Parameters
        ----------
        manager : Any
            The manager instance.
        path : str
            The path to the GRIB file.
        """
        self.manager = manager
        self.path = path

    def template(self, grib: str, lookup: Dict[str, Any]) -> ekd.Field:
        """Retrieve the template from the GRIB file.

        Parameters
        ----------
        grib : str
            The GRIB string.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        ekd.Field
            The field from the GRIB file.
        """
        return ekd.from_source("file", self.path)[0]



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from abc import ABC

LOG = logging.getLogger(__name__)


class Task(ABC):
    """Abstract base class for tasks.

    Parameters
    ----------
    name : str
        The name of the task.
    """

    def __init__(self, name: str) -> None:
        """Initialize the Task.

        Parameters
        ----------
        name : str
            The name of the task.
        """
        self.name = name

    def __repr__(self) -> str:
        """Return a string representation of the Task.

        Returns
        -------
        str
            String representation of the Task.
        """
        return f"{self.__class__.__name__}({self.name})"



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from typing import Dict

from anemoi.utils.registry import Registry

from anemoi.inference.config import Configuration
from anemoi.inference.task import Task
from anemoi.inference.transport import Transport

transport_registry = Registry(__name__)


def create_transport(config: Configuration, couplings: Configuration, tasks: Dict[str, Task]) -> Transport:
    """Create a transport instance based on the given configuration.

    Parameters
    ----------
    config : Configuration
        The configuration for the transport.
    couplings : Any
        The couplings for the transport.
    tasks : Any
        The tasks to be executed.

    Returns
    -------
    Any
        The created transport instance.
    """
    return transport_registry.from_config(config, couplings, tasks)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from argparse import ArgumentParser
from argparse import Namespace
from copy import deepcopy

from . import Command

LOG = logging.getLogger(__name__)


class SanitiseCmd(Command):
    """Sanitise a checkpoint file."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("path", help="Path to the checkpoint.")

    def run(self, args: Namespace) -> None:
        """Run the sanitise command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        from anemoi.utils.checkpoints import load_metadata
        from anemoi.utils.checkpoints import replace_metadata
        from anemoi.utils.sanitise import sanitise

        original_metadata, supporting_arrays = load_metadata(args.path, supporting_arrays=True)
        metadata = deepcopy(original_metadata)
        metadata = sanitise(metadata)

        if metadata != original_metadata:
            LOG.info("Patching metadata")
            assert "sources" in metadata["dataset"]
            replace_metadata(args.path, metadata, supporting_arrays)
        else:
            LOG.info("Metadata is already sanitised")


command = SanitiseCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from functools import wraps
from typing import Any
from typing import Callable

from anemoi.inference.context import Context

MARKER = object()


class main_argument:
    """Decorator to set the main argument of a function.

    For example...

    @main_argument("path")
    def grib_file_output(context, path, encoding=None, archive_requests=None):
        ...

    So we can have:

    output:
        grib: out.grib

    means the same as

    output:
        grib:
            path: out.grib
    """

    def __init__(self, name: str):
        """Initialize the main_argument decorator.

        Parameters
        ----------
        name : str
            The name of the main argument.
        """
        self.name = name

    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        """Decorate the function to set the main argument.

        Parameters
        ----------
        func : Callable[..., Any]
            The function to decorate.

        Returns
        -------
        Callable[..., Any]
            The decorated function.
        """

        @wraps(func)
        def wrapped(context: Context, main: object = MARKER, *args: Any, **kwargs: Any) -> Any:
            """Decorator function to set the main argument.

            Parameters
            ----------
            context : Any
                The context in which the function is called.
            main : Any, optional
                The main argument value, by default MARKER.
            *args : tuple
                Additional positional arguments.
            **kwargs : dict
                Additional keyword arguments.

            Returns
            -------
            Any
                The result of the decorated function.
            """
            if main is not MARKER:
                kwargs[self.name] = main

            return func(context, *args, **kwargs)

        return wrapped
